# This program is in the public domain
# Author Paul Kienzle
"""
Reflectometry materials.

Materials (see :class:`Material`) have a composition and a density.
Density may not be known, either because it has not been measured or
because the measurement of the bulk value does not apply to thin films.
The density parameter can be fitted directly, or the bulk density can be
used, and a stretch parameter can be fitted.

Mixtures (see :class:`Mixture`) are a special kind of material which are
composed of individual parts in proportion.  A mixture can be constructed
in a number of ways, such as by measuring proportional masses and mixing
or measuring proportional volumes and mixing.  The parameter of interest
may also be the relative number of atoms of one material versus another.
The fractions of the different mixture components are fitted parameters,
with the remainder of the bulk filled by the final component.

SLDs (see :class:`SLD`) are raw scattering length density values.  These
should be used if the material composition is not known.  In that case,
you will need separate SLD materials for each wavelength and probe.

*air* (see :class:`Vacuum`) is a predefined scatterer transparent to
all probes.

Scatter (see :class:`Scatterer`) is the abstract base class from which
all scatterers are derived.

The probe cache (see :class:`ProbeCache`) stores the scattering factors
for the various materials and calls the material sld method on demand.
Because the same material can be used for multiple measurements, the
scattering factors cannot be stored with material itself, nor does it
make sense to store them with the probe.  The scattering factor lookup
for the material is separate from the scattering length density
calculation so that you only need to look up the material once per fit.

The probe itself deals with all computations relating to the radiation
type and energy.  Unlike the normally tabulated scattering factors f', f''
for X-ray, there is no need to scale by probe by electron radius.  In
the end, sld is just the returned scattering factors times density.
"""

__all__ = [
    "Material",
    "Mixture",
    "SLD",
    "Vacuum",
    "Scatterer",
    "ProbeCache",
    "BulkDensityMaterial",
    "NaturalDensityMaterial",
    "NumberDensityMaterial",
    "RelativeDensityMaterial",
    "CellVolumeMaterial",
]

from dataclasses import dataclass, field
from typing import Optional, Any, Union, Dict, Callable, Literal, Tuple, List, Literal

import numpy as np
from numpy import inf, nan
import periodictable
from periodictable.constants import avogadro_number
from bumps.parameter import Expression, Parameter, to_dict  # , PARAMETER_TYPES, UnaryExpression
from periodictable.formulas import Formula as BaseFormula


class Scatterer:
    """
    A generic scatterer separates the lookup of the scattering factors
    from the calculation of the scattering length density.  This allows
    programs to fit density and alloy composition more efficiently.

    .. Note::
       the Scatterer base class is extended by
       :class:`_MaterialStacker <refl1d.model._MaterialStacker>` so that materials
       can be implicitly converted to slabs when used in stack construction
       expressions. It is not done directly to avoid circular dependencies
       between :mod:`model <refl1d.model>` and :mod:`material <refl1d.material>`.
    """

    def sld(self, sf) -> Tuple[float, float]:
        """
        Return the scattering length density pair (rho, irho) expected for the given
        scattering factors, as returned from a call to scattering_factors()
        for a particular probe.
        """
        raise NotImplementedError()

    def __or__(self, other):
        """
        Interface for a material stacker, to support e.g., *stack = Si | air*.
        """
        raise NotImplementedError("failed monkey-patch: material stacker needs" " to replace __or__ in Scatterer")

    def __call__(self, *args, **kw):
        """
        Interface for a material stacker, to support e.g., *stack = Si(thickness=)*.
        """
        raise NotImplementedError("failed monkey-patch: material stacker needs" " to replace __call__ in Scatterer")

    def __str__(self):
        return self.name


# ============================ No scatterer =============================
@dataclass(init=False)
class Vacuum(Scatterer):
    """
    Empty layer (all sld = 0)
    """

    def __init__(self, *args, **kw):
        self.name = "Vacuum"

    def parameters(self):
        return []

    def to_dict(self):
        return {
            "type": type(self).__name__,
        }

    def sld(self, probe):
        return 0, 0

    def __repr__(self):
        return "Vacuum()"


# ============================ Unknown scatterer ========================
@dataclass(init=False)
class SLD(Scatterer):
    r"""
    Unknown composition.

    Use this when you don't know the composition of the sample.  The
    absorption and scattering length density are stored directly rather
    than trying to guess at the composition from details about the sample.

    The complex scattering potential is defined by $\rho + j \rho_i$.
    Note that this differs from $\rho + j \mu/(2 \lambda)$ more
    traditionally used in neutron reflectometry, and $N r_e (f_1 + j f_2)$
    traditionally used in X-ray reflectometry.

    Given that $f_1$ and $f_2$ are always wavelength dependent for X-ray
    reflectometry, it will not make much sense to uses this for wavelength
    varying X-ray measurements.  Similarly, some isotopes, particularly
    rare earths, show wavelength dependence for neutrons, and so
    time-of-flight measurements should not be fit with a fixed SLD scatterer.
    """

    name: str
    rho: Parameter
    irho: Parameter

    def __init__(self, name="SLD", rho: Union[float, Parameter] = 0, irho=0):
        self.name = name
        self.rho = Parameter.default(rho, name=name + " rho")
        self.irho = Parameter.default(irho, name=name + " irho")

    def parameters(self):
        return {"rho": self.rho, "irho": self.irho}

    def to_dict(self):
        return to_dict(
            {
                "type": type(self).__name__,
                "name": self.name,
                "rho": self.rho,
                "irho": self.irho,
            }
        )

    def sld(self, probe):
        return self.rho.value, self.irho.value


# ============================ Substances =============================
# TODO: fix and use the real (circular) type definition:
# FormulaStructure = Tuple['FormulaUnit']
# FormulaUnit = Tuple[float, Union[str, FormulaStructure]]
# still have to convert Element to/from string repr
# ELEMENTS = Enum('elements', [(e.name, e.symbol) for e in periodictable.elements._element.values()])


class BaseMaterial(Scatterer):
    """
    Description of a solid block of material.

    :Parameters:
        *name* : string

        *formula* : Formula

            Composition can be initialized from either a string or a chemical
            formula.  Valid values are defined in periodictable.formula.

        *use_incoherent* = False : boolean

            True if incoherent scattering should be interpreted as absorption.

    """

    name: str
    formula: str  # Formula
    use_incoherent: bool = False

    # not in the schema:
    _formula: BaseFormula

    def __init__(
        self, formula: Union[str, BaseFormula], density=None, natural_density=None, name=None, use_incoherent=False
    ):
        # coerce formula to string repr:
        self.formula = str(formula)
        self._formula: BaseFormula = periodictable.formula(formula, density=density, natural_density=natural_density)
        self.name = name if name is not None else str(self._formula)
        self.use_incoherent = use_incoherent

    def sld(self, probe):
        rho, irho, incoh = probe.scattering_factors(self._formula, density=self.density.value)
        if self.use_incoherent:
            raise NotImplementedError("incoherent scattering not supported")
            # irho += incoh
        return rho, irho

    def __str__(self):
        return self.name


FitByChoices = Literal["bulk_density", "natural_density", "relative_density", "number_density", "cell_volume"]


def Material(
    formula: Union[str, BaseFormula],
    density: Optional[float] = None,
    natural_density: Optional[float] = None,
    name: Optional[str] = None,
    use_incoherent: bool = False,
    fitby: FitByChoices = "bulk_density",
    **kw,
):
    if fitby == "bulk_density":
        return BulkDensityMaterial(formula, density=density, name=name, use_incoherent=use_incoherent)
    elif fitby == "natural_density":
        return NaturalDensityMaterial(
            formula, natural_density=natural_density, name=name, use_incoherent=use_incoherent
        )
    elif fitby == "relative_density":
        return RelativeDensityMaterial(
            formula, density=density, natural_density=natural_density, name=name, use_incoherent=use_incoherent, **kw
        )
    elif fitby == "number_density":
        return NumberDensityMaterial(
            formula, density=density, natural_density=natural_density, name=name, use_incoherent=use_incoherent, **kw
        )
    elif fitby == "cell_volume":
        return CellVolumeMaterial(
            formula, density=density, natural_density=natural_density, name=name, use_incoherent=use_incoherent, **kw
        )
    else:
        raise ValueError(
            f'unknown value "{fitby}" of fitby: should be one of ["bulk_density", "natural_density", "relative_density", "number_density", "cell_volume"]'
        )


@dataclass(init=False)
class BulkDensityMaterial(BaseMaterial):
    """
    A solid block of material, described by its bulk density

    :Parameters:
        *density* : float | |g/cm^3|
            the bulk density for the material.
    """

    name: str
    formula: str  # Formula
    use_incoherent: bool = False
    density: Parameter

    def __init__(
        self,
        formula: Union[str, BaseFormula],
        density: Optional[Union[float, Parameter]] = None,
        name=None,
        use_incoherent=False,
    ):
        BaseMaterial.__init__(self, formula, density=density, name=name, use_incoherent=use_incoherent)
        if density is None:
            if self._formula.density is not None:
                density = self._formula.density
            else:
                raise ValueError(
                    f"material {self._formula} does not have known density: please provide it in arguments"
                )
        self.density = Parameter.default(density, name=self.name + " density", limits=(0, inf))
        self.bulk_density = self.density

    def parameters(self):
        return dict(density=self.density)


@dataclass(init=False)
class NaturalDensityMaterial(BaseMaterial):
    """
    A solid block of material, described by its natural density

    :Parameters:
        *natural_density* : float | |g/cm^3|
            the natural bulk density for the material.
    """

    name: str
    formula: str  # Formula
    use_incoherent: bool = False
    natural_density: Parameter
    density: Expression

    def __init__(
        self,
        formula: Union[str, BaseFormula],
        natural_density: Optional[Union[float, Parameter]] = None,
        name=None,
        use_incoherent=False,
    ):
        BaseMaterial.__init__(self, formula, natural_density=natural_density, name=name, use_incoherent=use_incoherent)
        if natural_density is None:
            if self._formula.density is not None:
                natural_density = self._formula.density
            else:
                raise ValueError(
                    f"material {self._formula} does not have known natural_density: please provide it in arguments"
                )
        self.natural_density = Parameter.default(natural_density, name=self.name + " nat. density", limits=(0, inf))
        self.density = self.natural_density / self._formula.natural_mass_ratio()

    def parameters(self):
        return dict(density=self.density, natural_density=self.natural_density)


@dataclass(init=False)
class NumberDensityMaterial(BaseMaterial):
    """
    A solid block of material, described by its number density

    :Parameters:
        *number_density*: [atoms/cm^3]
                Density is *number_density* * molar mass / avogadro constant

        *density* : float | |g/cm^3|
            if specified, the bulk density for the material.

        *natural_density* : float | |g/cm^3|
            if specified, the natural bulk density for the material.
    """

    name: str
    formula: str  # Formula
    use_incoherent: bool = False
    number_density: Parameter
    density: Expression

    def __init__(
        self,
        formula: Union[str, BaseFormula],
        number_density: Optional[Union[float, Parameter]] = None,
        density: Optional[float] = None,
        natural_density: Optional[float] = None,
        name=None,
        use_incoherent=False,
    ):
        BaseMaterial.__init__(
            self, formula, density=density, natural_density=natural_density, name=name, use_incoherent=use_incoherent
        )
        if number_density is None:
            if self._formula.density is not None:
                number_density = avogadro_number / self._formula.mass * self._formula.density
            else:
                raise ValueError(
                    f"material {self._formula} does not have known density: please provide density or natural_density argument"
                )
        self.number_density = Parameter.default(number_density, name=self.name + " number density", limits=(0, inf))
        self.density = self.number_density / avogadro_number * self._formula.mass
        self._formula.density = float(self.density)

    def parameters(self):
        return dict(density=self.density, number_density=self.number_density)


@dataclass(init=False)
class RelativeDensityMaterial(BaseMaterial):
    """
    A solid block of material, described by its relative density

    :Parameters:
        *relative_density* : unitless
            Density is *relative_density* * formula density

        *density* : float | |g/cm^3|
            if specified, the bulk density for the material.

        *natural_density* : float | |g/cm^3|
            if specified, the natural bulk density for the material.
    """

    name: str
    formula: str  # Formula
    use_incoherent: bool = False
    relative_density: Parameter
    density: Expression

    def __init__(
        self,
        formula: Union[str, BaseFormula],
        relative_density: Optional[Union[float, Parameter]] = None,
        density: Optional[float] = None,
        natural_density: Optional[float] = None,
        name=None,
        use_incoherent=False,
    ):
        BaseMaterial.__init__(
            self, formula, density=density, natural_density=natural_density, name=name, use_incoherent=use_incoherent
        )
        if self._formula.density is None and density is None and natural_density is None:
            raise ValueError(
                f"material {self._formula} does not have known density: please provide density or natural_density argument"
            )
        if relative_density is None:
            relative_density = 1
        self.relative_density = Parameter.default(relative_density, name=self.name + " rel. density", limits=(0, inf))
        self.density = self._formula.density * self.relative_density

    def parameters(self):
        return dict(density=self.density, relative_density=self.relative_density)


@dataclass(init=False)
class CellVolumeMaterial(BaseMaterial):
    """
    A solid block of material, described by the volume of one unit cell

    :Parameters:
        *cell_volume* : |Ang^3|
                Density is mass / *cell_volume*

        *density* : float | |g/cm^3|
            if specified, the bulk density for the material.

        *natural_density* : float | |g/cm^3|
            if specified, the natural bulk density for the material.


    For example, to fit Pd by cell volume use::
        >>> m = Material('Pd', fitby='cell_volume')
        >>> m.cell_volume.range(1, 10)
        Parameter(Pd cell volume)
        >>> print("%.2f %.2f"%(m.density.value, m.cell_volume.value))
        12.02 14.70

    """

    name: str
    formula: str  # Formula
    use_incoherent: bool = False
    cell_volume: Parameter
    density: Expression

    def __init__(
        self,
        formula: Union[str, BaseFormula],
        cell_volume: Optional[Union[float, Parameter]] = None,
        density: Optional[float] = None,
        natural_density: Optional[float] = None,
        name=None,
        use_incoherent=False,
    ):
        BaseMaterial.__init__(
            self, formula, density=density, natural_density=natural_density, name=name, use_incoherent=use_incoherent
        )
        if self._formula.density is None and density is None and natural_density is None:
            raise ValueError(
                f"material {self._formula} does not have known density: please provide density or natural_density argument"
            )
        if cell_volume is None:
            cell_volume = (1e24 * self._formula.molecular_mass) / self._formula.density
        self.cell_volume = Parameter.default(cell_volume, name=self.name + " cell volume", limits=(0, inf))
        self.density = (1e24 * self._formula.molecular_mass) / self.cell_volume

    def parameters(self):
        return dict(density=self.density, cell_volume=self.cell_volume)


MaterialTypes = Union[
    BulkDensityMaterial, NaturalDensityMaterial, NumberDensityMaterial, RelativeDensityMaterial, CellVolumeMaterial
]


class Compound(Scatterer):
    """
    Chemical formula with variable composition.

    :Parameters:

        *parts* : [M1, F1, M2, F2, ...]

    Unlike a simple material which has a chemical formula and a density,
    the formula itself can be varied by fitting the number of atoms of
    each component in the unit cell in addition to the overall density.

    An individual component can be a chemical formula, not just an element.
    """

    def __init__(self, parts=None):
        # Split [M1, N1, M2, N2, ...] into [M1, M2, ...], [N1, N2, ...]
        formula = [parts[i] for i in range(0, len(parts), 2)]
        count = [parts[i] for i in range(1, len(parts), 2)]
        # Convert M1, M2, ... to materials if necessary
        formula = [periodictable.formula(p) for p in formula]
        count = [Parameter.default(w, limits=(0, None), name=str(f) + " count") for w, f in zip(count, formula)]
        self.parts = formula
        self.count = count

    def parameters(self):
        """
        Adjustable parameters are the fractions associated with each
        constituent and the relative scale fraction used to tweak
        the overall density.
        """
        return {"count": self.count}

    def to_dict(self):
        return {
            "type": type(self).__name__,
            "parts": to_dict(self.parts),
            "count": to_dict(self.count),
        }

    def formula(self):
        return tuple((c.value, f) for c, f in zip(self.count, self.parts))

    def __str__(self):
        return "<%s>" % (", ".join(str(M) for M in self.formula()))

    def __repr__(self):
        return "Compound([%s])" % (", ".join(repr(M) for M in self.formula()))


# ============================ Alloys =============================


class _VolumeFraction(object):
    """
    Returns the relative volume for each component in the system given
    the volume percentages.
    """

    def __init__(self, base, material):
        pass

    def __call__(self, fraction):
        return 0.01 * np.asarray(fraction)


class _MassFraction(object):
    """
    Returns the relative volume for each component in the system given
    the relative masses.
    """

    def __init__(self, base, material):
        self._material = [base] + material

    def __call__(self, fraction):
        density = np.array([m.density.value for m in self._material])
        volume = fraction / density
        return volume / sum(volume)


class Mixture(Scatterer):
    """
    Mixed block of material.

    The components of the mixture can vary relative to each other, either
    by mass, by volume or by number::

        Mixture.bymass(base, M2, F2..., name='mixture name')
        Mixture.byvolume(base, M2, F2..., name='mixture name')

    The materials *base*, *M2*, *M3*, ... can be chemical formula
    strings including @density or from material objects. Use natural_density
    to change from bulk values if the formula has isotope substitutions.

    The fractions F2, F3, ... are percentages in [0, 100]. The implicit
    fraction for the base material is 100 - (F2+F3+...). The SLD is NaN
    then *F1 < 0*.

    name defaults to M2.name+M3.name+...
    """

    @classmethod
    def bymass(cls, base, *parts, **kw):
        """
        Returns an alloy defined by relative mass of the constituents.

        Mixture.bymass(base, M2, F2, ..., name='mixture name')
        """
        return cls(base, parts, by="mass", **kw)

    @classmethod
    def byvolume(cls, base, *parts, **kw):
        """
        Returns an alloy defined by relative volume of the constituents.

        Mixture.byvolume(base, M2, F2, ..., name='mixture name')
        """
        return cls(base, parts, by="volume", **kw)

    def __init__(self, base, parts, by="volume", name=None, use_incoherent=False):
        # Split [M1, M2, F2, ...] into [M1, M2, ...], [F2, ...]
        material = [parts[i] for i in range(0, len(parts), 2)]
        fraction = [parts[i] for i in range(1, len(parts), 2)]
        # Convert M1, M2, ... to materials if necessary
        if not isinstance(base, Scatterer):
            base = Material(base)
        material = [p if isinstance(p, Scatterer) else Material(p) for p in material]

        # Specify the volume calculator based on the type of fraction
        if by == "volume":
            _volume = _VolumeFraction(base, material)
        elif by == "mass":
            _volume = _MassFraction(base, material)
        else:
            raise ValueError("fraction must be one of volume, mass or number")

        # Name defaults to names of individual components
        if name is None:
            name = "+".join(p.name for p in material)

        # Make the fractions into fittable parameters
        fraction = [Parameter.default(w, limits=(0, 100), name=f.name + "% " + by) for w, f in zip(fraction, material)]

        self._volume = _volume
        self.base = base
        self.material = material
        self.fraction = fraction
        self.name = name
        self.use_incoherent = use_incoherent

    def parameters(self):
        """
        Adjustable parameters are the fractions associated with each
        constituent and the relative scale fraction used to tweak
        the overall density.
        """
        return {
            "base": self.base.parameters(),
            "material": [m.parameters() for m in self.material],
            "fraction": self.fraction,
        }

    def to_dict(self):
        return {
            "type": type(self).__name__,
            "base": to_dict(self.base),
            "material": to_dict(self.material),
            "fraction": to_dict(self.fraction),
        }

    def _density(self):
        """
        Compute the density of the mixture from the density and proportion
        of the individual components.
        """
        fraction = np.array([0.0] + [m.value for m in self.fraction])
        # TODO: handle invalid fractions using penalty functions
        # S = sum(fraction)
        # scale = S/100 if S > 100 else 1
        # fraction[0] = 100 - S/scale
        # penalty = scale - 1
        fraction[0] = 100 - sum(fraction)
        if (fraction < 0).any():
            return nan
        volume = self._volume(fraction)
        density = np.array([m.density() for m in [self.base] + self.material])
        return np.sum(volume * density)

    density = property(_density, doc=_density.__doc__)

    def sld(self, probe):
        """
        Return the scattering length density and absorption of the mixture.
        """
        # Convert fractions into an array, with the final fraction
        fraction = np.array([0.0] + [f.value for f in self.fraction])
        fraction[0] = 100 - sum(fraction)
        # TODO: handle invalid fractions using penalty functions
        # S = sum(fraction)
        # scale = S/100 if S > 100 else 1
        # fraction[0] = 100 - S/scale
        # penalty = scale - 1
        if (fraction < 0).any():
            return nan, nan

        # Lookup SLD
        slds = [c.sld(probe) for c in [self.base] + self.material]
        rho, irho = [np.asarray(v) for v in zip(*slds)]

        # Use calculator to convert individual SLDs to overall SLD
        volume_fraction = self._volume(fraction)
        rho = np.sum(rho * extend(volume_fraction, rho))

        irho = np.sum(irho * extend(volume_fraction, irho))
        if self.use_incoherent:
            raise NotImplementedError("incoherent scattering not supported")
        # print "Mixture", self.name, coh, absorp

        return rho, irho

    def __str__(self):
        return "<%s>" % (", ".join(str(M) for M in [self.base] + self.material))

    def __repr__(self):
        return "Mixture(%s)" % (", ".join(repr(M) for M in [self.base] + self.material))


# ============================ SLD cache =============================


class ProbeCache(object):
    """
    Probe proxy for materials properties.

    A caching probe which only looks up scattering factors for materials
    which it hasn't seen before.   Note that caching is based on object
    id, and will fail if the material object is updated with a new atomic
    structure.

    *probe* is the probe to use when looking up the scattering length density.

    The scattering factors need to be retrieved each time the probe
    or the composition changes. This can be done either by deleting
    an individual material from probe (using del probe[material]) or
    by clearing the entire cash.
    """

    def __init__(self, probe=None):
        self._probe = probe
        self._cache = {}

    def clear(self):
        self._cache = {}

    def __delitem__(self, material):
        if material in self._cache:
            del self._cache[material]

    def scattering_factors(self, material, density):
        """
        Return the scattering factors for the material, retrieving them from
        the cache if they have already been looked up.
        """
        h = id(material)
        if h not in self._cache:
            # lookup density of 1, and scale to actual density on retrieval
            self._cache[h] = self._probe.scattering_factors(material, density=1.0)
        return [v * density for v in self._cache[h]]


def extend(a, b):
    """
    Extend *a* to match the number of dimensions of *b*.

    This adds dimensions to the end of *a* rather than the beginning. It is
    equivalent to *a[..., None, None]* with the right number of None elements
    to make the number of dimensions match (or np.newaxis if you prefer).

    For example::

        from numpy.random import rand
        a, b = rand(3, 4), rand(3, 4, 2)
        a + b
        ==> ValueError: operands could not be broadcast together with shapes (3,4) (3,4,2)
        c = extend(a, b) + b
        c.shape
        ==> (3, 4, 2)

    Numpy broadcasting rules automatically extend arrays to the beginning,
    so the corresponding *lextend* function is not needed::

        c = rand(3, 4) + rand(2, 3, 4)
        c.shape
        ==> (2, 3, 4)
    """
    if np.isscalar(a):
        return a
    # CRUFT: python 2.7 support
    extra_dims = (1,) * (b.ndim - a.ndim)
    return a.reshape(a.shape + extra_dims)
    # python 3 uses
    # extra_dims = (np.newaxis,)*(b.ndim-a.ndim)
    # return a[(..., *extra_dims)]



# This program is in the public domain
# Author: Paul Kienzle
r"""
Read and write staj files

Staj files are the model files for the mlayer and gj2 programs, which are
used as the calculation engine for the reflpak suite. Mlayer supports
unpolarized beam with multilayer models,  and has files ending in
**.staj**. GJ2 supports polarized beam without multilayer models, and
has files ending in **.sta**.
"""

from math import pi

import numpy as np
from bumps.wsolve import wsolve

ERF_FWHM = 2.35482004503095  # 2 * sqrt(2*log(2))
TANH_FWHM = 0.47320111770856327  # 1/2 atanh(erf(1/sqrt(2))) / acosh(sqrt(2))
# Derivation
# ==========
# Converting from error function FWHM to 1-sigma requires solving
# exp(-0.5*(z/1)**2) = 0.5 for z and doubling it, giving 2*sqrt(2*log(2))
#
# Converting from tanh FWHM to 1-sigma error function is more complicated.
# First find C where w is defined as 1-sigma equivalent of tanh, using the
# identity Erf.CDF(z=sigma;w=sigma) = tanh.CDF(z=sigma;w=sigma).
# This simplifies to::
#
#    erf.CDF  = (1+erf(z/(w*sqrt(2)))/2 = (1+erf(1/sqrt(2)))/2
#    tanh.CDF = (1+tanh(C/w*z))/2       = (1+tanh(C))/2
#
#    erf.CDF = tanh.CDF => C = atanh(erf(1/sqrt(2)))
#
# Next find C where w is defined as FWHM, using the equivalent probability
# density function::
#
#    PDF(z) = C/2w * sech(C/w*z)**2
#
# Solving PDF(w/2) = PDF(0)/2 yields::
#
#    Pw = PDF(w/2) = C/2w * sech(C/2)**2
#    Po = PDF(0) = C/2w * sech(0)**2/2 = C/2w
#
#    Pw = Po/2 => sech(C/2)**2 = 1/2
#              => C = 2 acosh(sqrt(2))
#
# To find 1-sigma width given tanh FWHM of w, use the scale factor
# s = C_1_sigma/C_fwhm = 1/2 atanh(erf(1/sqrt(2)))/acosh(sqrt(2))


# ----------------------------------------------------------------------------
# The format of a Non-Magnetic Staj file is shown below.  It is an ASCII text
# file composed of lines as follows:
#
# Note that nLayers = nTLayers + nMLayers + nBLayers + 3
# This is the number of top, middle, and bottom layers plus extra profile lines.
#
# Note that nRepeat is the number of repeats of the mLayers profile lines.
# This parameter is ignored by KsRefl.
#
# Line 1:         nTLayers  nMLayers  nBLayers  nRepeat  nFitParams  nRoughSteps
# Line 2:         wavelength  wavelengthDiv  angularDiv  [theta_offset]
# Line 3:         intensity  background  Qmin  Qmax  nQ (data points)
# Line 4:         profileType ('E' for error function, 'H' for tanh)
# Line 5:         datafileName
# Line 6:         <optional outputfileName>
##Sections have an ignored layer followed by the layers for the section
##The ignored layer of the top section contains vacuum SLD
##Layers have rho  mrho  depth  rough  mu
# Line 7 to 7+nL: sections
# Line 7+nL+1:    p1 p2 p3 ... (fitted parameter numbers)
# Line 7+nL+2 to end: constraints
## Constraints often end with a line of garbage characters.
#
# On reading a Staj file the following conversions are made:
# - scattering_length_density = 1e6 rho / (16 pi)
# - magnetic_scat_len_density = 1e6 mrho / (16 pi)
# - roughness_between_layers  = rough
# - absorption                = mu / (2 wavelength)
#
# -------------------------------------------------------------------------------
# The following shows the contents of an sample Non-Magnetic Staj file:
#            1            1            1            1            0           21
#        6.00000       0.160000    0.000300000       0.000000
#        1.00000   1.00000E-010     0.00741408       0.109617            267
# E
# SNS_Ni_Si_3col.txt
#
#  0.000000E+000  0.000000E+000  0.000000E+000  1.000000E-010  0.000000E+000
#  0.000000E+000  0.000000E+000  9.000000E+002  0.000000E+000  0.000000E+000
#  0.000000E+000  0.000000E+000  0.000000E+000  1.000000E-010  0.000000E+000
#  4.452130E-004  0.000000E+000  7.439860E+002  5.308710E+001  0.000000E+000
#  4.452130E-004  0.000000E+000  0.000000E+000  5.308710E+001  0.000000E+000
#  1.279860E-004  0.000000E+000  6.250000E+002  1.814250E+001  0.000000E+000
#
# MQC1=MQC2
# ]
#
# -------------------------------------------------------------------------------
# The sample Non-Magnetic Staj File shown above (after conversion of data
# elements) consists of the following layers:
#
# - T0 (Air)  rho =  0.0, depth =   0.0, mu = 0.0, mrho = 0.0
#                                          roughness between layers =  0.0
# - T1 (Air)  rho =  0.0, depth = 900.0, mu = 0.0, mrho = 0.0
#                                          roughness between layers = 22.5
# - M1 (Ni)   rho = 8.86, depth = 744.0, mu = 0.0, mrho = 0.0
#                                          roughness between layers = 7.70
# - B1 (Si)   rho = 2.55, depth = 625.0, mu = 0.0, mrho = 0.0
#
# Note that comments or additional blank lines are not permitted except after
# all required lines (i.e. line 1 through line 6+nLayers).
#
# Note that the number of values on the first line determines if the format is
# for a non-magnetic (6 values) or a magnetic (4 values) Staj file.


class MlayerModel(object):
    r"""
    Model definition used by MLayer program.

    **Attributes:**

    Q values and reflectivity come from a data file with Q, R, dR or
    from simulation with linear spacing from Qmin to Qmax in equal steps:

        *data_file*
            name of the data file, or None if this is simulation only
        *Qmin*, *Qmax*, *num_Q*
            for simulation, Q sample points

    Resolution is defined by wavelength and by incident angle:

        *wavelength*, *wavelength_dispersion*, *angular_divergence*
            resolution is calculated as
            $\Delta Q/Q = \Delta\lambda/\lambda + \Delta\theta/\theta$

    Additional beam parameters correct for intensity, background and
    possibly sample alignment:

        *intensity*, *background*
            incident beam intensity and sample background
        *theta_offset*
            alignment angle correction

    The model is defined in terms of layers, with three sections.  The top
    and bottom section correspond to the fixed layers at the surface and
    the substrate.  The middle section layers can be repeated an arbitrary
    number of times, as defined by the number of repeats attribute.  The
    attributes defining the sections are:

        *num_top* *num_middle* *num_bottom*
            section sizes
        *num_repeats*
            number of times middle section repeats

    Interfaces are split into discrete steps according to a profile,
    either error function or hyperbolic tangent.  For sharp interfaces
    which do not overlap within a layer, the interface is broken into a
    fixed number of slabs with slabs having different widths, but equal
    changes in height.  For broad interfaces, the whole layer is split
    into the same fixed number of slabs, but with each slab having the
    same width.  The following attributes are used:

        *roughness_steps*
            number of roughness steps (13 is coarse; 51 is fine)
        *roughness_profile*
            roughness profile is either 'E' for error function or 'H' for tanh

    Layers have thickness, interface roughness and real and imaginary
    scattering length density (SLD).  Roughness is stored in the file
    using full width at half maximum (FWHM) for the given profile type.
    For convenience, roughness can also be set or queried using a 1-\ $\sigma$
    equivalent roughness on an error function profile.  Regardless,
    layer parameters are represented as vectors with one entry for each
    top, middle and bottom layer using the following attributes:

        *thickness*, *roughness* : float | |Ang|
            layer thickness and FWHM roughness
        *rho*, *irho*, *incoh* : float | |1e-6/Ang^2|
            complex coherent $\rho + j \rho_i$ and incoherent SLD

    Computed attributes are provided for convenience:

        *sigma_roughness* : float | |Ang|
            1-\ $\sigma$ equivalent roughness for erf profile
        *mu*
            absorption cross section (2*wavelength*irho + incoh)

    .. Note::
          The staj files store SLD as $16\pi\rho$, $2\lambda\rho_i$
          with an additional column of 0 for magnetic SLD. This conversion
          happens automatically on read/write. The incoherent cross section
          is assumed to be zero.

    The layers are ordered from surface to substrate.

    Additional attributes are as follows:

        *fitpars*
            individual fit parameter numbers
        *constraints*
            constraints between layers
        *output_file*
            name of the output file

    These can be safely ignored, except perhaps if you want to try to
    compile the constraints into something that can be used by your system.

    **Methods:**

    model = MlayerModel(attribute=value, ...)

        Construct a new MLayer model with the given attributes set.

    model = MlayerModel.load(filename)

        Construct a new MLayer model from a staj file.

    model.set(attribute=value, ...)

        Replace a set of attribute values.

    model.fit_resolution(Q, dQ)

        Choose the best resolution parameters to match the given Q, dQ
        resolution.  Returns the object so that calls can be chained.

    model.resolution(Q)

        Return the resolution at Q for the current resolution parameters.

    model.split_sections()

        Assign top, middle, bottom and repeats to distribute the layers
        across sections.  Returns the object so that calls can be chained.

    model.save(filename)

        Write the model to the given named file.  Raises ValueError if
        the model is invalid.

    **Constructing new files:**

    Staj files can be constructed directly.  The MlayerModel constructor
    can accept all data attributes as key word arguments.  Models require
    at least *data_file*, *wavelength*, *thickness*, *roughness* and *rho*.
    Resolution parameters can be set using model.fit_resolution(Q, dQ).
    Section sizes can be set using model.split_sections().  Everything
    else has reasonable defaults.

    """

    data_file = ""
    Qmin = 0
    Qmax = 0.5
    num_Q = 200
    wavelength = 1
    wavelength_dispersion = 0.01
    angular_divergence = 0.001
    intensity = 1
    background = 0
    theta_offset = 0
    num_top, num_middle, num_bottom = 0, 0, 0
    num_repeats = 1
    roughness_steps = 13
    roughness_profile = "E"
    thickness = roughness = rho = None
    irho = incoh = 0
    fitpars = []
    constraints = ""
    output_file = ""

    def __init__(self, **kw):
        self.set(**kw)

    def set(self, **kw):
        valid = (
            "data_file",
            "Qmin",
            "Qmax",
            "num_Q",
            "wavelength",
            "wavelength_dispersion",
            "angular_divergence",
            "intensity",
            "background",
            "theta_offset",
            "num_top",
            "num_middle",
            "num_bottom",
            "num_repeats",
            "roughness_steps",
            "roughness_profile",
            "sigma_roughness",
            "thickness",
            "roughness",
            "rho",
            "irho",
            "incoh",
            "fitpars",
            "constraints",
            "output_file",
        )
        for k, v in kw.items():
            if k not in valid:
                raise TypeError("Unexpected attribute '%s' in Mlayer Model" % k)
            setattr(self, k, v)

    @classmethod
    def load(cls, filename):
        """
        Load a staj file, returning an MlayerModel object
        """
        fin = open(filename, "r")
        lines = fin.readlines()
        fin.close()

        self = cls()
        self._parse(lines)
        return self

    def save(self, filename):
        """
        Save the staj file
        """
        self._check()
        fid = open(filename, "w")
        try:
            self._write(fid)
        finally:
            fid.close()

    def FWHMresolution(self, Q):
        r"""
        Return the resolution at Q for mlayer with the current settings
        for wavelength, wavelength divergence and angular divergence.

        Resolution is full-width at half maximum (FWHM), not 1-\ $\sigma$.
        """
        return (abs(Q) * self.wavelength_dispersion + 4 * pi * self.angular_divergence) / self.wavelength

    def fit_FWHMresolution(self, Q, dQ, weight=1):
        r"""
        Choose the best dL and dT to match the resolution dQ.

        Given that mlayer uses the following resolution function:

        .. math::

            \Delta Q_k = (|Q_k| \Delta\lambda + 4 \pi \Delta\theta)/\lambda_k

        we can use a linear system solver to find the optimal
        $\Delta \lambda$ and $\Delta \theta$ across our dataset from the
        over-determined system:

        .. math::

          [|Q_k|/\lambda_k, 4\pi/\lambda_k][\Delta\lambda, \Delta\theta]^T
              = \Delta Q_k

        If weights are provided (e.g., $\Delta R_k/R_k$), then weigh each
        point during the fit.

        Given that the experiment is often run with fixed slits at the
        start and end, you may choose to match the resolution across the
        entire $Q$ range, or instead restrict it to just the region where
        the slits are opening.  You will generally want to get the resolution
        correct at the critical edge since that's where it will have the
        largest effect on the fit.

        Returns the object so that operations can be chained.
        """
        A = np.array([abs(Q) / self.wavelength, np.ones_like(Q) * (4 * pi / self.wavelength)]).T
        s = wsolve(A, y=dQ, dy=weight)
        self.wavelength_dispersion = s.x[0]
        self.angular_divergence = s.x[1]

        return self

    def split_sections(self):
        """
        Split the given set of layers into sections, putting as many layers
        as possible into the middle section, then the bottom and finally
        the top.

        Returns the object so that operations can be chained.
        """
        self.num_repeats = 1
        n = len(self.thickness)
        if n > 28:
            raise ValueError("A maximum of 28 layers is allowed")
        if n < 2:
            raise ValueError("Must have at least two layers")
        n -= 1  # Incident medium layer
        if n >= 11:
            self.num_middle = 9
        elif n > 2:
            self.num_middle = n - 2
        elif n > 1:
            self.num_middle = 1
        else:
            self.num_middle = 0
        n -= self.num_middle
        if n >= 10:
            self.num_bottom = 9
        else:
            self.num_bottom = n - 1
        n -= self.num_bottom
        self.num_top = n  # This may be zero if there are less than 3 layers.

        return self

    def __str__(self):
        line = []
        if self.data_file != "":
            line.append("Data: %s" % self.data_file)
        else:
            line.append("Q: %g to %g in %d steps" % (self.Qmin, self.Qmax, self.num_Q))
        line.append(
            "Wavelength L: %g  dL/L: %g, dTheta: %g"
            % (self.wavelength, self.wavelength_dispersion, self.angular_divergence)
        )
        line.append(
            "Beam intensity: %g  Background: %g, Theta offset: %g"
            % (self.intensity, self.background, self.theta_offset)
        )
        profile = "error function" if self.roughness_profile == "E" else "tanh"
        line.append("Interface: %s in %d steps" % (profile, self.roughness_steps))
        w, s = self.thickness, self.roughness
        rho, mu = self.rho, self.mu
        line.append("Layers:")
        line.append(("   " + "%15s " * 4) % ("Width (A)", "Interface (FWHM)", "Rho (1e-6/A)", "Mu (1e-6/A)"))
        for i in range(len(self.rho)):
            if i == 0:
                name = "V"
            elif i < self.num_top + 1:
                name = "T%d" % (i)
            elif i < self.num_top + self.num_middle + 1:
                name = "M%d" % (i - self.num_top)
            else:
                name = "B%d" % (i - self.num_top - self.num_middle)
            line.append(("%s:" + ("%15g " * 4)) % (name, w[i], s[i], rho[i], mu[i]))
        if self.constraints != "":
            line.append("Constraints:")
            line.append(self.constraints)
        return "\n".join(line)

    def _check(self):
        """
        Verify that the staj file is correct and ready for writing, filling
        in the details that are missing.
        """
        if (
            (self.irho is not None and len(self.rho) != len(self.irho))
            or len(self.rho) != len(self.thickness)
            or len(self.rho) != len(self.roughness)
        ):
            raise ValueError("layer parameters have different lengths")

        # Could check if Qmin/Qmax/num_Q matches data, but I don't think
        # it matters, so skip it

        if self.num_top + self.num_middle + self.num_bottom + 1 != len(self.rho):
            raise ValueError("section sizes do not match number of layers")

    def _get_mu(self):
        mu = 2 * self.wavelength * self.irho + self.incoh
        return mu * np.ones_like(self.rho)

    def _set_mu(self, mu):
        self.irho = mu / (2 * self.wavelength)

    mu = property(fget=_get_mu, fset=_set_mu)

    def _get_sigma(self):
        if self.roughness_profile == "H":
            scale = TANH_FWHM
        else:
            scale = ERF_FWHM
        return self.roughness / scale

    def _set_sigma(self, v):
        if self.roughness_profile == "H":
            scale = TANH_FWHM
        else:
            scale = ERF_FWHM
        self.roughness = v * scale

    sigma_roughness = property(fget=_get_sigma, fset=_set_sigma)

    def _parse(self, lines):
        # 1: num_top num_middle num_bottom num_repeats num_fit rough_steps
        nums = [int(s) for s in lines[0].split()]
        self.num_top, self.num_middle, self.num_bottom, self.num_repeats, _, self.roughness_steps = nums

        # 2: wavelength  wavelength_dispersion  angular_divergence [theta_offset]
        nums = [float(s) for s in lines[1].split()]
        self.wavelength, self.wavelength_dispersion, self.angular_divergence = nums[:3]
        self.theta_offset = nums[3] if len(nums) > 3 else 0

        # 3: intensity  background  Qmin  Qmax  num_Q
        nums = [float(s) for s in lines[2].split()]
        self.intensity, self.background, self.Qmin, self.Qmax = nums[:4]
        self.num_Q = int(nums[4])

        # 4: profile_type ('E' for error function, 'H' for tanh)
        self.roughness_profile = lines[3].strip().upper()[:1]
        if self.roughness_profile not in ("E", "H"):
            raise ValueError("Expected roughness profile type E or H")

        # 5: data_file
        # 6: output_file
        self.data_file = lines[4].strip()
        self.output_file = lines[5].strip()

        # nL = num_top+num_middle+num_bottom+3
        # 7 to 7+nL: rho mrho depth rough mu
        # ignore the layer before each section
        nL = self.num_top + self.num_middle + self.num_bottom + 3
        layers = [[float(v) for v in line.split()] for line in lines[6 : 6 + nL]]
        del layers[self.num_top + 1]
        del layers[self.num_top + self.num_middle + 1]

        A = np.array(layers)
        self.rho = A[:, 0] * (1e6 / 16 / pi)
        self.irho = A[:, 4] * (1e6 / 2 / self.wavelength)
        self.incoh = A[:, 0] * 0
        self.thickness = A[:, 2]
        self.roughness = A[:, 3]

        # 7+nL+1: P1 P2 P3 ...  (fit parameters)
        self.fitpars = [int(s) for s in lines[6 + nL].split()]
        # 7+nL+2 to end: constraints
        self.constraints = "".join(lines[6 + nL + 1 :])

    def _write(self, fid):
        # 1: num_top num_middle num_bottom num_repeats num_fit rough_steps
        fid.write(
            "%d %d %d %d %d %d\n"
            % (
                self.num_top,
                self.num_middle,
                self.num_bottom,
                self.num_repeats,
                len(self.fitpars),
                self.roughness_steps,
            )
        )

        # 2: wavelength  wavelength_dispersion  angular_divergence [theta_offset]
        fid.write(
            "%g %g %g %g\n" % (self.wavelength, self.wavelength_dispersion, self.angular_divergence, self.theta_offset)
        )

        # 3: intensity  background  Qmin  Qmax  num_Q
        fid.write("%g %g %g %g %d\n" % (self.intensity, self.background, self.Qmin, self.Qmax, self.num_Q))

        # 4: profile_type ('E' for error function, 'H' for tanh)
        fid.write("%s\n" % self.roughness_profile)

        # 5: data_file
        # 6: output_file
        fid.write("%s\n%s\n" % (self.data_file, self.output_file))

        # nL = num_top+num_middle+num_bottom+3
        # 7 to 7+nL: rho mrho depth rough mu
        # ignore the layer before each section
        rho = self.rho * (16 * pi * 1e-6)
        mu = self.mu * 1e-6
        w, s = self.thickness, self.roughness

        def _write_layer(idx):
            fid.write("%g %g %g %g %g\n" % (rho[idx], 0.0, w[idx], s[idx], mu[idx]))

        offset = 0
        for n in [self.num_top, self.num_middle, self.num_bottom]:
            _write_layer(offset)
            if n == 0:
                # In the case of only two or three layers, some sections
                # may have no layers, and need to be filled with a repeated
                # value from the next section.
                _write_layer(offset)
            for i in range(n):
                _write_layer(i + offset + 1)
            offset += n
        # 7+nL+1: P1 P2 P3 ...  (fit parameters)
        fid.write(" ".join(str(p) for p in self.fitpars) + "\n")
        # 7+nL+2 to end: constraints
        fid.write(self.constraints)


# ==============================================================================
# Translate the Staj file into model representation (magnetic case).
# ==============================================================================


# -------------------------------------------------------------------------------
# The format of a Magnetic Staj file is shown below.  It is an ASCII text file
# composed of lines as follows:
#
# Note that nLayers does not count the top layer, just the middle and bottom
# layers (though data for the top (incident) layer is provided) in the file.
#
# Line 1:         wavelength  wavelengthDiv  angularDiv  [aguide]
# Line 2:         intensity  background
# Line 3:         nLayers  nRoughSteps  nFitParam
# Line 4:         Qmin  Qmax  nQ (data points in 'a' datafle)
# Line 5:         Qmin  Qmax  nQ (data points in 'b' datafle)
# Line 6:         Qmin  Qmax  nQ (data points in 'c' datafle)
# Line 7:         Qmin  Qmax  nQ (data points in 'd' datafle)
# Line 8:         profileType ('E' for error function, 'H' for tanh)
# Line 9:         active cross sections (usually 'abcd' or 'ABCD')
# Line 10:        data file (base name without suffix char such as test.refl)
# Line 11:        output file
## Layer information follows in sets of 3 lines x (nL+1).
# Line next (1):  rho   depth   rough  mu
# Line next (2):  mrho  mdepth  mrough (mrho is also known as phi)
# Line next (3):  mtheta
## Fit information fills the remainder of the file
# Line 11+3*(nL+1)+1:  Fit parameters (integers)
# Line 11+3*(nL+1)+2 to end-of-file:  Constraint program
#
# On reading a Staj file the following conversions are made:
# - scattering length density = rho * (1e6 / 16 / pi )
# - complex SLD               = irho * (1e6 / 2 / wavelength)
# - magnetic SLD              = mrho * (1e6 / 16 / pi )
# - structural roughness      = rough
# - magnetic roughness        = mrough
#
# -------------------------------------------------------------------------------
# The following shows the contents of an sample Magnetic Staj file:
#
#        5.00000      0.0500000   1.00000E-005       -90.0000
#        1.00000   1.00000E-010
#            7            7            0
#     0.00613985       0.174845            215
#     0.00613985       0.174845            215
#     0.00613985       0.174845            215
#     0.00613985       0.174845            215
# E
# abcd
# Test.refl
#
#       0.000000       0.000000       0.000000       0.000000
#       0.000000       0.000000       0.000000
#        270.000
#    0.000318683        50.0000        2.00000       0.000000
#       0.000000        50.0000        2.00000
#        270.000
#    0.000329239        80.0000        2.00000       0.000000
#       0.000000        80.0000        2.00000
#        270.000
#    0.000352864        216.000        2.00000       0.000000
#    0.000117621        216.000        2.00000
#        270.000
#    0.000329239        80.0000        2.00000       0.000000
#       0.000000        80.0000        2.00000
#        270.000
#    0.000318683        50.0000        2.00000       0.000000
#       0.000000        50.0000        2.00000
#        270.000
#   9.70124E-005        10.0000        2.00000       0.000000
#       0.000000        10.0000        2.00000
#        270.000
#    0.000104050        100.000        2.00000   1.00000E-009
#       0.000000        100.000        2.00000
#        270.000
#
# -------------------------------------------------------------------------------
# The sample Magnetic Staj File shown above (after conversion of data elements)
# consists of the following layers (with mdepth and mrough not shown):
#
# - MV (Air)   rho =  0.0, depth =   0.0, mu = 0.0, mrho =  0.0, mtheta = 270.0
#                                           roughness between layers =   0.0
# - M1 (Pt)    rho = 6.34, depth =  50.0, mu = 0.0, mrho =  0.0, mtheta = 270.0
#                                           roughness between layers = 0.849
# - M2 (Cu)    rho = 6.55, depth =  80.0, mu = 0.0, mrho =  0.0, mtheta = 270.0
#                                           roughness between layers = 0.849
# - M3 (Co+Ni) rho = 7.02, depth = 216.0, mu = 0.0, mrho = 2.34, mtheta = 270.0
#                                           roughness between layers = 0.849
# - M4 (Cu)    rho = 6.55, depth =  80.0, mu = 0.0, mrho =  0.0, mtheta = 270.0
#                                           roughness between layers = 0.849
# - M5 (Pt)    rho = 6.34, depth =  50.0, mu = 0.0, mrho =  0.0, mtheta = 270.0
#                                           roughness between layers = 0.849
# - M6 (SiO2)  rho = 1.93, depth =  10.0, mu = 0.0, mrho =  0.0, mtheta = 270.0
#                                           roughness between layers = 0.849
# - M7 (Si)    rho = 2.07, depth = 100.0, mu = 0.0, mrho =  0.0, mtheta = 270.0
#
# Note that comments or additional blank lines are not permitted except after
# all required lines (i.e. line 1 through line 11+3*(nLayers+1)).
#
# Note that the number of values on the first line determines if the format is
# for a non-magnetic (6 values) or a magnetic (4 values) Staj file.
#
class MlayerMagnetic(object):
    r"""
    Model definition used by GJ2 program.

    **Attributes:**

    Q values and reflectivity come from a data file with Q, R, dR or
    from simulation with linear spacing from Qmin to Qmax in equal steps:

        *data_file*
            base name of the data file, or None if this is simulation only
        *active_xsec*
            active cross sections (usually 'abcd' for all cross sections)
        *Qmin*, *Qmax*, *num_Q*
            for simulation, Q sample points

    Resolution is defined by wavelength and by incident angle:

        *wavelength*, *wavelength_dispersion*, *angular_divergence*
            resolution is calculated as
            $\Delta Q/Q = \Delta\lambda/\lambda + \Delta\theta/\theta$

    Additional beam parameters correct for intensity, background and
    possibly guide field angle:

        *intensity*, *background*
            incident beam intensity and sample background
        *guide_angle*
            angle of the guide field

    Unlike pure structural models, magnetic models are in one large
    section with no repeats.  The single parameter is the number of
    layers, which is implicit in the length of the layer data and
    does not need to be an explicit attribute.

    Interfaces are split into discrete steps according to a profile,
    either error function or hyperbolic tangent.  For sharp interfaces
    which do not overlap within a layer, the interface is broken into a
    fixed number of slabs with slabs having different widths, but equal
    changes in height.  For broad interfaces, the whole layer is split
    into the same fixed number of slabs, but with each slab having the
    same width. The following attributes are used:

        *roughness_steps*
            number of roughness steps (13 is coarse; 51 is fine)
        *roughness_profile*
            roughness profile is either 'E' for error function or 'H' for tanh

    Layers have thickness, interface roughness and real and imaginary
    scattering length density (SLD).  Roughness is stored in the file
    using full width at half maximum (FWHM) for the given profile type.
    For convenience, roughness can also be set or queried using a 1-\ $\sigma$
    equivalent roughness on an error function profile.  Regardless,
    layer parameters are represented as vectors with one entry for each
    top, middle and bottom layer using the following attributes:

        *thickness*, *roughness* : float | |Ang|
            layer thickness and FWHM roughness
        *rho*, *irho* : float, float | $16 \pi \rho$, $2\lambda\rho_i$
            complex scattering length density
        *mthickness*, *mroughness* : float | |Ang|
            magnetic thickness and roughness
        *mrho* : float | $16 \pi \rho_M$
            magnetic scattering length density
        *mtheta* : float | |deg|
            magnetic angle
        *sigma_roughness*, *sigma_mroughness* : float | |Ang|
            computed 1-\ $\sigma$ equivalent roughness for erf profile

    The conversion from stored $16 \pi \rho$, $2\lambda\rho_i$ to
    in memory $10^6 \rho$, $10^6 \rho_i$  happens automatically on
    read/write.

    The layers are ordered from surface to substrate.

    Additional attributes are as follows:

        *fitpars*
            individual fit parameter numbers

        *constraints*
            constraints between layers

        *output_file*
            name of the output file

    These can be safely ignored, except perhaps if you want to try to
    compile the constraints into something that can be used by your system.

    **Methods:**

    model = MlayerMagnetic(attribute=value, ...)

        Construct a new MLayer model with the given attributes set.

    model = MlayerMagnetic.load(filename)

        Construct a new MLayer model from a sta file.

    model.set(attribute=value, ...)

        Replace a set of attribute values.

    model.fit_resolution(Q, dQ)

        Choose the best resolution parameters to match the given Q, dQ
        resolution.  Returns the object so that calls can be chained.

    model.resolution(Q)

        Return the resolution at Q for the current resolution parameters.

    model.save(filename)

        Write the model to the given named file.  Raises ValueError if
        the model is invalid.

    **Constructing new files:**

    Staj files can be constructed directly.  The MlayerModel constructor
    can accept all data attributes as key word arguments.  Models require
    at least *data_file*, *wavelength*, *thickness*, *roughness* and *rho*.
    Resolution parameters can be set using model.fit_resolution(Q, dQ).
    Everything else has reasonable defaults.

    """

    data_file = ""
    active_xsec = "abcd"
    Qmin = 0
    Qmax = 0.5
    num_Q = 200
    wavelength = 1
    wavelength_dispersion = 0.01
    angular_divergence = 0.001
    intensity = 1
    background = 0
    guide_angle = 270
    roughness_steps = 13
    roughness_profile = "E"
    thickness = roughness = rho = irho = None
    mthickness = mroughness = mrho = mtheta = None
    fitpars = []
    constraints = ""
    output_file = ""

    def __init__(self, **kw):
        self.set(**kw)

    def set(self, **kw):
        valid = (
            "data_file",
            "active_xsec",
            "Qmin",
            "Qmax",
            "num_Q",
            "wavelength",
            "wavelength_dispersion",
            "angular_divergence",
            "intensity",
            "background",
            "guide_angle",
            "roughness_steps",
            "roughness_profile",
            "thickness",
            "roughness",
            "rho",
            "irho",
            "sigma_roughness",
            "mthickness",
            "mroughness",
            "mrho",
            "mtheta",
            "sigma_mroughness",
            "fitpars",
            "constraints",
            "output_file",
        )
        for k, v in kw.items():
            if k not in valid:
                raise TypeError("Unexpected attribute '%s' in Mlayer Model" % k)
            setattr(self, k, v)

    @classmethod
    def load(cls, filename):
        """
        Load a staj file, returning an MlayerModel object
        """
        fin = open(filename, "r")
        lines = fin.readlines()
        fin.close()

        self = cls()
        try:
            self._parse(lines)
        except:
            raise ValueError("Improper staj file")
        return self

    def save(self, filename):
        """
        Save the staj file
        """
        self._check()
        fid = open(filename, "w")
        try:
            self._write(fid)
        finally:
            fid.close()

    def FWHMresolution(self, Q):
        return (abs(Q) * self.wavelength_dispersion + 4 * pi * self.angular_divergence) / self.wavelength

    FWHMresolution.__doc__ = MlayerModel.FWHMresolution.__doc__

    def fit_FWHMresolution(self, Q, dQ, weight=1):
        A = np.array([abs(Q) / self.wavelength, np.ones_like(Q) * (4 * pi / self.wavelength)])
        s = wsolve(A, y=dQ, dy=weight)
        self.wavelength_dispersion = s.x[0]
        self.angular_divergence = s.x[1]

        return self

    fit_FWHMresolution.__doc__ = MlayerModel.fit_FWHMresolution.__doc__

    def __str__(self):
        line = []
        if self.data_file != "":
            line.append("Data: %s[%s]" % (self.data_file, self.active_xsec.upper()))
        else:
            line.append("Q: %g to %g in %d steps" % (self.Qmin, self.Qmax, self.num_Q))
        line.append(
            "Wavelength L: %g  dL/L: %g, dTheta: %g"
            % (self.wavelength, self.wavelength_dispersion, self.angular_divergence)
        )
        line.append(
            "Beam intensity: %g  Background: %g, Guide angle: %g" % (self.intensity, self.background, self.guide_angle)
        )
        profile = "error function" if self.roughness_profile == "E" else "tanh"
        line.append("Interface: %s in %d steps" % (profile, self.roughness_steps))
        w, s = self.thickness, self.roughness
        wm = self.mthickness if self.mthickness is not None else w
        sm = self.mroughness if self.mroughness is not None else s
        rho = self.rho
        irho = self.irho if self.irho is not None else np.zeros_like(w)
        mrho = self.mrho if self.mrho is not None else np.zeros_like(w)
        mtheta = self.mtheta if self.mtheta is not None else 270 * np.ones_like(w)
        line.append("Layers:")
        line.append(
            ("    " + ("%15s " * 4) + "\n    " + ("%15s " * 4))
            % (
                "Width (A)",
                "Interface (FWHM)",
                "Rho (1e-6/A)",
                "iRho (1e-6/A)",
                "Mag width",
                "Mag interface",
                "Mag rho",
                "Mag angle (deg)",
            )
        )
        for i in range(len(self.rho)):
            line.append(
                ("%3d:" + ("%15g " * 4) + "\n    " + ("%15g " * 4))
                % (i, w[i], s[i], rho[i], irho[i], wm[i], sm[i], mrho[i], mtheta[i])
            )
        if self.constraints != "":
            line.append("Constraints:")
            line.append(self.constraints)
        return "\n".join(line)

    def _check(self):
        """
        Verify that the staj file is correct and ready for writing, filling
        in the details that are missing.
        """
        ns = [
            (len(v) if v is not None else 0)
            for v in (
                self.rho,
                self.thickness,
                self.roughness,
                self.irho,
                self.mrho,
                self.mthickness,
                self.mroughness,
                self.mtheta,
            )
        ]
        if any((n != ns[0] and n != 0) for n in ns[1:]):
            raise ValueError("layer parameters have different lengths")
        if any((n == 0) for n in ns[:3]):
            raise ValueError("rho, thickness and roughness are required")

        # Could check if Qmin/Qmax/num_Q matches data, but I don't think
        # it matters, so skip it

    def _get_sigma(self):
        if self.roughness_profile == "H":
            scale = TANH_FWHM
        else:
            scale = ERF_FWHM
        return self.roughness / scale

    def _set_sigma(self, v):
        if self.roughness_profile == "H":
            scale = TANH_FWHM
        else:
            scale = ERF_FWHM
        self.roughness = v * scale

    sigma_roughness = property(_get_sigma, _set_sigma)

    def _get_msigma(self):
        if self.roughness_profile == "H":
            scale = TANH_FWHM
        else:
            scale = ERF_FWHM
        return self.mroughness / scale

    def _set_msigma(self, v):
        if self.mroughness_profile == "H":
            scale = TANH_FWHM
        else:
            scale = ERF_FWHM
        self.mroughness = v * scale

    sigma_mroughness = property(_get_msigma, _set_msigma)

    def _parse(self, lines):
        # 1: wavelength  wavelength_dispersion  angular_divergence [aguide]
        nums = [float(s) for s in lines[0].split()]
        self.wavelength, self.wavelength_dispersion, self.angular_divergence = nums[:3]
        self.guide_angle = nums[3] if len(nums) > 3 else 0

        # 2: intensity  background
        nums = [float(s) for s in lines[1].split()]
        self.intensity, self.background = nums

        # 3: maxLayer  nRoughSteps  nFitParam
        nums = [int(s) for s in lines[2].split()]
        maxLayer, self.roughness_steps, _ = nums

        # 4-7: Qmin  Qmax  nQ (data points in a, b, c and d)
        # Note that we are only keeping the first one; for simulation all the
        # others should be the same, and for loading, the datafile will tell
        # us what Q to use.
        nums = [float(s) for s in lines[3].split()]
        self.Qmin, self.Qmax, self.num_Q = nums[0], nums[1], int(nums[2])

        # 8: profile_type ('E' for error function, 'H' for tanh)
        self.roughness_profile = lines[7].strip()
        if self.roughness_profile not in ("E", "H"):
            raise ValueError("Expected roughness profile type E or H")

        # 9: active cross sections (usually 'abcd' or 'ABCD')
        self.active_xsec = lines[8].strip().lower()

        # 10: data file (base name without suffix char such as test.refl)
        # 11: output_file
        self.data_file = lines[9].strip()
        self.output_file = lines[10].strip()

        ## Layer information follows in sets of 3 lines x (nL+1).
        # Line next (1):  rho   depth   rough  mu
        # Line next (2):  mrho  mdepth  mrough (mrho is also known as phi)
        # Line next (3):  mtheta
        nL = maxLayer + 1
        layers = [[float(v) for v in " ".join(lines[11 + 3 * i : 11 + 3 * (i + 1)]).split()] for i in range(nL)]

        A = np.array(layers)
        self.rho = A[:, 0] * (1e6 / 16 / pi)
        self.irho = A[:, 3] * (1e6 / 2 / self.wavelength)
        self.thickness = A[:, 1]
        self.roughness = A[:, 2]
        self.mrho = A[:, 4] * (1e6 / 16 / pi)
        self.mtheta = A[:, 7]
        self.mthickness = A[:, 5]
        self.mroughness = A[:, 6]

        ## Fit information fills the remainder of the file
        # footer+1: P1 P2 P3 ...  (fit parameters)
        self.fitpars = [int(s) for s in lines[10 + 3 * nL + 1].split()]
        # footer+2 to end: constraints
        self.constraints = "".join(lines[10 + 3 * nL + 2 :])

    def _write(self, fid):
        # 1: wavelength  wavelength_dispersion  angular_divergence [aguide]
        fid.write(
            "%g %g %g %g\n" % (self.wavelength, self.wavelength_dispersion, self.angular_divergence, self.guide_angle)
        )

        # 2: intensity  background
        fid.write("%g %g\n" % (self.intensity, self.background))

        # 3: maxLayer  nRoughSteps  nFitParam
        fid.write("%d %d %d\n" % (len(self.rho) - 1, self.roughness_steps, len(self.fitpars)))

        # 4-7: Qmin  Qmax  nQ (data points in a, b, c and d)
        fid.write("%g %g %d\n" % (self.Qmin, self.Qmax, self.num_Q))
        fid.write("%g %g %d\n" % (self.Qmin, self.Qmax, self.num_Q))
        fid.write("%g %g %d\n" % (self.Qmin, self.Qmax, self.num_Q))
        fid.write("%g %g %d\n" % (self.Qmin, self.Qmax, self.num_Q))

        # 8: profile_type ('E' for error function, 'H' for tanh)
        fid.write("%s\n" % self.roughness_profile)

        # 9: active cross sections (usually 'abcd' or 'ABCD')
        fid.write(" %s\n" % self.active_xsec)

        # 10: data file (base name without suffix char such as test.refl)
        # 11: output_file
        fid.write("%s\n%s\n" % (self.data_file, self.output_file))

        ## Layer information follows in sets of 3 lines x (nL+1).
        # Line next (1):  rho   depth   rough  mu
        # Line next (2):  mrho  mdepth  mrough (mrho is also known as phi)
        # Line next (3):  mtheta
        rho = self.rho * (16 * pi / 1e6)
        w, s = self.thickness, self.roughness
        wm = self.mthickness if self.mthickness is not None else self.thickness
        sm = self.mroughness if self.mroughness is not None else self.roughness
        if self.irho is not None:
            mu = self.irho * (2 * self.wavelength / 1e6)
        else:
            mu = np.zeros_like(rho)
        if self.mrho is not None:
            mrho = self.mrho * (16 * pi / 1e6)
        else:
            mrho = np.zeros_like(rho)
        if self.mtheta is not None:
            mtheta = self.mtheta
        else:
            mtheta = 270 * np.ones_like(rho)
        for i in range(len(rho)):
            fid.write("%g %g %g %g\n%g %g %g\n%g\n" % (rho[i], w[i], s[i], mu[i], mrho[i], wm[i], sm[i], mtheta[i]))

        ## Fit information fills the remainder of the file
        # footer+1: P1 P2 P3 ...  (fit parameters)
        fid.write(" ".join(str(p) for p in self.fitpars) + "\n")
        # footer+2 to end: constraints
        fid.write(self.constraints)



from dataclasses import dataclass
from typing import List
from bumps.fitproblem import FitProblem as _FitProblem
from .experiment import Experiment

# @dataclass(init=False)
# class BaseFitProblem(_BaseFitProblem):
#     fitness: Experiment


@dataclass(init=False)
class FitProblem(_FitProblem):
    __doc__ = _FitProblem.__doc__.replace("Fitness", "Experiment")
    models: List[Experiment]



"""
Monotonic spline modeling for free interfaces
"""

from __future__ import division, with_statement

from dataclasses import dataclass, field
from typing import Optional, Any, Union, Dict, Callable, Literal, Tuple, List, Literal
import numpy as np
from numpy import diff, hstack, sqrt, searchsorted, asarray, cumsum, inf, nonzero, linspace, sort, isnan, clip
from bumps.parameter import Parameter as Par, Function as ParFunction, to_dict, Constant
from bumps.mono import monospline, count_inflections

from . import util
from .model import Layer


# TODO: add left_sld, right_sld to all layers so that fresnel works
# TODO: access left_sld, right_sld so freeform doesn't need left, right
# TODO: restructure to use vector parameters
# TODO: allow the number of layers to be adjusted by the fit
class FreeLayer(Layer):
    """
    A freeform section of the sample modeled with splines.

    sld (rho) and imaginary sld (irho) can be modeled with a separate
    number of control points. The control points can be equally spaced
    in the layers unless rhoz or irhoz are specified. If the z values
    are given, they must be in the range [0, 1].  One control point is
    anchored at either end, so there are two fewer z values than controls
    if z values are given.

    Layers have a slope of zero at the ends, so the automatically blend
    with slabs.
    """

    def __init__(self, below=None, above=None, thickness=0, z=(), rho=(), irho=(), name="Freeform"):
        self.name = name
        self.below, self.above = below, above
        self.thickness = Par.default(thickness, name=name + " thickness", limits=(0, inf))
        self.interface = Constant(0, name=name + " interface")

        def parvec(vector, name, limits):
            return [Par.default(p, name=name + "[%d]" % i, limits=limits) for i, p in enumerate(vector)]

        self.rho, self.irho, self.z = [
            parvec(v, name + " " + part, limits)
            for v, part, limits in zip((rho, irho, z), ("rho", "irho", "z"), ((-inf, inf), (0, inf), (0, 1)))
        ]
        if len(self.z) != len(self.rho):
            raise ValueError("must have one z for each rho value")
        if len(self.irho) > 0 and len(self.z) != len(self.irho):
            raise ValueError("must have one z for each irho value")

    def parameters(self):
        return {
            "thickness": self.thickness,
            "rho": self.rho,
            "irho": self.irho,
            "z": self.z,
            "below": self.below.parameters(),
            "above": self.above.parameters(),
        }

    def to_dict(self):
        return to_dict(
            {
                "type": type(self).__name__,
                "name": self.name,
                "parameters": self.parameters(),
            }
        )

    def profile(self, Pz, below, above):
        thickness = self.thickness.value
        rbelow, ibelow = below
        rabove, iabove = above
        z = sort([0] + [p.value for p in self.z] + [1]) * thickness

        rho = hstack((rbelow, [p.value for p in self.rho], rabove))
        Prho = monospline(z, rho, Pz)

        if len(self.irho) > 0:
            irho = hstack((ibelow, [p.value for p in self.irho], iabove))
            Pirho = monospline(z, irho, Pz)
        else:
            Pirho = 0 * Prho
        return Prho, Pirho

    def penalty(self):
        dz = diff([p.value for p in self.z])
        return np.sum(dz[dz < 0] ** 2)

    def render(self, probe, slabs):
        below = self.below.sld(probe)
        above = self.above.sld(probe)
        Pw, Pz = slabs.microslabs(self.thickness.value)
        Prho, Pirho = self.profile(Pz, below, above)
        slabs.extend(rho=[Prho], irho=[Pirho], w=Pw)


def inflections(dx, dy):
    x = hstack((0, cumsum(dx)))
    y = hstack((0, cumsum(dy)))
    return count_inflections(x, y)


@dataclass(init=False)
class FreeInterface(Layer):
    """
    A freeform section of the sample modeled with monotonic splines.

    Layers have a slope of zero at the ends, so the automatically blend
    with slabs.
    """

    name: Optional[str]
    below: Optional[Any]
    above: Optional[Any]
    thickness: Par
    interface: Par
    dz: List[Union[float, Par]]
    dp: List[Union[float, Par]]
    # inflections: List[Any]

    def __init__(self, thickness=0, interface=0, below=None, above=None, dz=None, dp=None, name="Interface"):
        self.name = name
        self.below, self.above = below, above
        self.thickness = Par.default(thickness, limits=(0, inf), name=name + " thickness")
        self.interface = Constant(0, name=name + " interface")

        # Choose reasonable defaults if not given
        if dp is None and dz is None:
            dp = [1] * 5
        if dp is None:
            dp = [1] * len(dz)
        if dz is None:
            dz = [1] * len(dp)
        if len(dz) != len(dp):
            raise ValueError("Need one dz for every dp")

        # if len(z) != len(vf)+2:
        #    raise ValueError("Only need vf for interior z, so len(z)=len(vf)+2")
        self.dz = [Par.default(p, name=name + " dz[%d]" % i, limits=(0, inf)) for i, p in enumerate(dz)]
        self.dp = [Par.default(p, name=name + " dp[%d]" % i, limits=(0, inf)) for i, p in enumerate(dp)]
        self.inflections = Par(name=name + " inflections")
        self.inflections.equals(ParFunction(inflections, dx=self.dz, dy=self.dp))

    def parameters(self):
        return {
            "thickness": self.thickness,
            "interface": self.interface,
            "dz": self.dz,
            "dp": self.dp,
            "below": self.below.parameters(),
            "above": self.above.parameters(),
            "inflections": self.inflections,
        }

    def to_dict(self):
        pars = self.parameters()
        pars.pop("inflections")  # derived parameter not needed for save
        return to_dict(
            {
                "type": type(self).__name__,
                "name": self.name,
                "parameters": pars,
            }
        )

    def profile(self, Pz):
        thickness = self.thickness.value
        z, p = [hstack((0, cumsum(asarray([v.value for v in vector], "d")))) for vector in (self.dz, self.dp)]
        if p[-1] == 0:
            p[-1] = 1
        p *= 1 / p[-1]
        z *= thickness / z[-1]
        profile = clip(monospline(z, p, Pz), 0, 1)
        return profile

    def render(self, probe, slabs):
        thickness = self.thickness.value
        # interface ignored for FreeInterface
        # interface = self.interface.value
        below_rho, below_irho = self.below.sld(probe)
        above_rho, above_irho = self.above.sld(probe)
        # Pz is the center, Pw is the width
        Pw, Pz = slabs.microslabs(thickness)
        profile = self.profile(Pz)
        Pw, profile = util.merge_ends(Pw, profile, tol=1e-3)
        Prho = (1 - profile) * below_rho + profile * above_rho
        Pirho = (1 - profile) * below_irho + profile * above_irho
        slabs.extend(rho=[Prho], irho=[Pirho], w=Pw)


# CRUFT: still working on best rep'n for control point locations
class _FreeInterfaceW(Layer):
    """
    A freeform section of the sample modeled with monotonic splines.

    Layers have a slope of zero at the ends, so the automatically blend
    with slabs.
    """

    def __init__(self, interface=0, below=None, above=None, dz=None, dp=None, name="Interface"):
        self.name = name
        self.below, self.above = below, above
        self.interface = Par.default(interface, limits=(0, inf), name=name + " interface")

        # Choose reasonable defaults if not given
        if dp is None and dz is None:
            dp = [1] * 5
        if dp is None:
            dp = [1] * len(dz)
        if dz is None:
            dz = [10.0 / len(dp)] * len(dp)
        if len(dz) != len(dp):
            raise ValueError("Need one dz for every dp")
        # if len(z) != len(vf)+2:
        #    raise ValueError("Only need vf for interior z, so len(z)=len(vf)+2")
        self.dz = [Par.default(p, name=name + " dz[%d]" % i, limits=(0, inf)) for i, p in enumerate(dz)]
        self.dp = [Par.default(p, name=name + " dp[%d]" % i, limits=(0, inf)) for i, p in enumerate(dp)]

    def _get_thickness(self):
        w = sum(v.value for v in self.dz)
        return Par(w, name=self.name + " thickness")

    def _set_thickness(self, v):
        if v != 0:
            raise ValueError("thickness cannot be set for FreeformInterface")

    thickness = property(_get_thickness, _set_thickness)

    def parameters(self):
        return {
            "interface": self.interface,
            "dz": self.dz,
            "dp": self.dp,
            "below": self.below.parameters(),
            "above": self.above.parameters(),
        }

    def to_dict(self):
        pars = self.parameters()
        pars.pop("inflections")  # derived parameter not needed for save
        return to_dict(
            {
                "type": type(self).__name__,
                "name": self.name,
                "parameters": pars,
            }
        )

    def render(self, probe, slabs):
        interface = self.interface.value
        below_rho, below_irho = self.below.sld(probe)
        above_rho, above_irho = self.above.sld(probe)
        z = hstack((0, cumsum([v.value for v in self.dz])))
        p = hstack((0, cumsum([v.value for v in self.dp])))
        thickness = z[-1]
        if p[-1] == 0:
            p[-1] = 1
        p /= p[-1]
        Pw, Pz = slabs.microslabs(z[-1])
        profile = monospline(z, p, Pz)
        Pw, profile = util.merge_ends(Pw, profile, tol=1e-3)
        Prho = (1 - profile) * below_rho + profile * above_rho
        Pirho = (1 - profile) * below_irho + profile * above_irho
        slabs.extend(rho=[Prho], irho=[Pirho], w=Pw)



import math

Z_EPS = 1e-6


def align_magnetic(d, sigma, rho, irho, dM, sigmaM, rhoM, thetaM, output_flat):
    # ignoring thickness d on the first and last layers
    # ignoring interface width sigma on the last layer
    nlayers = len(d)
    nlayersM = len(dM)
    noutput = nlayers + nlayersM

    # making sure there are at least two layers
    if nlayers < 2 or nlayersM < 2:
        raise ValueError("only works with more than one layer")

    output = output_flat
    magnetic = 0  # current magnetic layer index
    nuclear = 0  # current nuclear layer index
    z = 0.0  # current interface depth
    next_z = 0.0  # next nuclear interface
    next_zM = 0.0  # next magnetic interface
    # int active = 3# active interfaces, active&0x1 for nuclear, active&0x2 for magnetic

    k = 0  # current output layer index
    while True:
        # repeat over all nuclear/magnetic layers
        if k == noutput:
            return -1  # exceeds capacity of output

        # printf("%d: %d %d %g %g %g\n", k, nuclear, magnetic, z, next_z, next_zM)
        # printf("%g %g %g %g\n", rho[nuclear], irho[nuclear], rhoM[magnetic], thetaM[magnetic])
        # printf("%g %g %g %g\n", d[nuclear], sigma[nuclear], dM[magnetic], sigmaM[magnetic])

        # Set the scattering strength using the current parameters
        output[k][2] = rho[nuclear]
        output[k][3] = irho[nuclear]
        output[k][4] = rhoM[magnetic]
        output[k][5] = thetaM[magnetic]

        # Check if we are at the last layer for both nuclear and magnetic
        # If so set thickness and interface width to zero.  We are doing a
        # center of the loop exit in order to make sure that the final layer
        # is added.

        if magnetic == nlayersM - 1 and nuclear == nlayers - 1:
            output[k][0] = 0.0
            output[k][1] = 0.0
            k += 1
            break

        # Determine if we are adding the nuclear or the magnetic interface next,
        # or possibly both.  The order of the conditions is important.
        #
        # Note: the final value for next_z/next_zM is not defined.  Rather than
        # checking if we are on the last layer we simply add the value of the
        # last thickness to z, which may be 0, nan, inf, or anything else.  This
        # doesn't affect the algorithm since we don't look at next_z when we are
        # on the final nuclear layer or next_zM when we are on the final magnetic
        # layer.
        #
        # Note: averaging nearly aligned interfaces can lead to negative thickness
        # Consider nuc = [1-a, 0, 1] and mag = [1+a, 1, 1] for 2a < Z_EPS.
        # On the first step we set next_z to 1-a, next_zM to 1+a and z to the
        # average of 1-a and 1+a, which is 1.  On the second step next_z is
        # still 1-a, so the thickness next_z - z = -a. Since a is tiny we can just
        # pretend that -a == zero by setting thickness to fmax(next_z - z, 0.0).

        if nuclear == nlayers - 1:
            # No more nuclear layers... play out the remaining magnetic layers.
            output[k][0] = max(next_zM - z, 0.0)
            output[k][1] = sigmaM[magnetic]
            magnetic += 1
            next_zM += dM[magnetic]
        elif magnetic == nlayersM - 1:
            # No more magnetic layers... play out the remaining nuclear layers.
            output[k][0] = max(next_z - z, 0.0)
            output[k][1] = sigma[nuclear]
            nuclear += 1
            next_z += d[nuclear]
        elif math.fabs(next_z - next_zM) < Z_EPS and math.fabs(sigma[nuclear] - sigmaM[magnetic]) < Z_EPS:
            # Matching nuclear/magnetic boundary, with almost identical interfaces.
            # Increment both nuclear and magnetic layers.
            output[k][0] = max(0.5 * (next_z + next_zM) - z, 0.0)
            output[k][1] = 0.5 * (sigma[nuclear] + sigmaM[magnetic])
            nuclear += 1
            next_z += d[nuclear]
            magnetic += 1
            next_zM += dM[magnetic]
        elif next_zM < next_z:
            # Magnetic boundary comes before nuclear boundary, so increment magnetic.
            output[k][0] = max(next_zM - z, 0.0)
            output[k][1] = sigmaM[magnetic]
            magnetic += 1
            next_zM += dM[magnetic]
        else:
            # Nuclear boundary comes before magnetic boundary
            # OR nuclear and magnetic boundaries match but interfaces are different.
            # so increment nuclear.
            output[k][0] = max(next_z - z, 0.0)
            output[k][1] = sigma[nuclear]
            nuclear += 1
            next_z += d[nuclear]

        z += output[k][0]
        k += 1

    return k


def contract_mag(d, sigma, rho, irho, rhoM, thetaM, dA):
    n = len(d)
    m = n - 1  # /* last middle layer */
    i = newi = 1  # /* Skip the substrate */
    while i < m:
        # /* Get ready for the next layer */
        # /* Accumulation of the first row happens in the inner loop */
        dz = 0
        rhoarea = irhoarea = rhoMpara_area = rhoMperp_area = thetaM_area = 0.0
        rholo = rhohi = rho[i]
        irholo = irhohi = irho[i]

        # /*
        #  * averaged thetaM is from atan2 (when rhoM is nonzero)
        #  * which returns values in the range -180 to 180,
        #  * so we keep track of the phase offset of the input
        #  * to match it afterward
        #  */
        thetaM_phase_offset = (thetaM[i] + 180.0) // 360.0

        # /* Pre-calculate projections */
        rhoM_sign = math.copysign(1, rhoM[i])
        thetaM_radians = thetaM[i] * math.pi / 180.0
        rhoMpara = rhoM[i] * math.cos(thetaM_radians)
        rhoMperp = rhoM[i] * math.sin(thetaM_radians)
        # /*
        #  * Note that mpara indicates M when theta_M = 0,
        #  * and mperp indicates M when theta_M = 90,
        #  * but in most cases M is parallel to H when theta_M = 270
        #  * and Aguide = 270
        #  */
        mparalo = mparahi = rhoMpara
        mperplo = mperphi = rhoMperp

        # /* Accumulate slices into layer */
        while i < m:
            # /* Accumulate next slice */
            dz += d[i]
            rhoarea += d[i] * rho[i]
            irhoarea += d[i] * irho[i]
            # /* thetaM_area is only used if rhoM is zero */
            thetaM_area += d[i] * thetaM[i]
            # /* Use pre-calculated next values */
            rhoMpara_area += rhoMpara * d[i]
            rhoMperp_area += rhoMperp * d[i]

            # /* If no more slices or sigma != 0, break immediately */
            i += 1
            if i == n or sigma[i - 1] != 0.0:
                break

            # /* If next slice exceeds limit then break */
            if rho[i] < rholo:
                rholo = rho[i]
            if rho[i] > rhohi:
                rhohi = rho[i]
            if (rhohi - rholo) * (dz + d[i]) > dA:
                break

            if irho[i] < irholo:
                irholo = irho[i]
            if irho[i] > irhohi:
                irhohi = irho[i]
            if (irhohi - irholo) * (dz + d[i]) > dA:
                break

            # /* Pre-calculate projections of next layer */
            thetaM_radians = thetaM[i] * math.pi / 180.0
            rhoMpara = rhoM[i] * math.cos(thetaM_radians)
            rhoMperp = rhoM[i] * math.sin(thetaM_radians)

            # /* If next slice is wrapped in phase, break */
            if (thetaM[i] + 180.0) // 360.0 != thetaM_phase_offset:
                break

            # /* If next slice has different sign for rhoM, break */
            if math.copysign(1, rhoM[i]) != rhoM_sign:
                break

            if rhoMpara < mparalo:
                mparalo = rhoMpara
            if rhoMpara > mparahi:
                mparahi = rhoMpara
            if (mparahi - mparalo) * (dz + d[i]) > dA:
                break

            if rhoMperp < mperplo:
                mperplo = rhoMperp
            if rhoMperp > mperphi:
                mperphi = rhoMperp
            if (mperphi - mperplo) * (dz + d[i]) > dA:
                break

        assert newi < m
        d[newi] = dz
        if dz == 0:
            rho[newi] = rho[i - 1]
            irho[newi] = irho[i - 1]
            rhoM[newi] = rhoM[i - 1]
            thetaM[newi] = thetaM[i - 1]
        else:
            rho[newi] = rhoarea / dz
            irho[newi] = irhoarea / dz
            mean_rhoMpara = rhoMpara_area / dz
            mean_rhoMperp = rhoMperp_area / dz
            mean_rhoM = math.sqrt(mean_rhoMpara**2 + mean_rhoMperp**2) * rhoM_sign
            if mean_rhoM == 0:
                # /* If rhoM is zero, then thetaM is meaningless: use plain average */
                thetaM[newi] = thetaM_area / dz
            else:
                # /* Otherwise, calculate the mean thetaM
                #  * invert the sign of components if rhoM is negative, before atan2 */
                thetaM_from_mean = math.atan2(mean_rhoMperp * rhoM_sign, mean_rhoMpara * rhoM_sign) * 180.0 / math.pi
                thetaM_from_mean += 360.0 * thetaM_phase_offset
                thetaM[newi] = thetaM_from_mean
            rhoM[newi] = mean_rhoM
        sigma[newi] = sigma[i - 1]

        newi += 1

    # /* Save the last layer */
    # /* Last layer uses surface values */
    rho[newi] = rho[n - 1]
    irho[newi] = irho[n - 1]
    rhoM[newi] = rhoM[n - 1]
    thetaM[newi] = thetaM[n - 1]
    # /* No interface for final layer */
    newi += 1

    return newi


def contract_by_area(d, sigma, rho, irho, dA):
    n = len(d)
    i = newi = 1  # /* Skip the substrate */
    while i < n:
        # /* Get ready for the next layer */
        # /* Accumulation of the first row happens in the inner loop */
        dz = rhoarea = irhoarea = 0.0
        rholo = rhohi = rho[i]
        irholo = irhohi = irho[i]

        # /* Accumulate slices into layer */
        while True:
            # /* Accumulate next slice */
            dz += d[i]
            rhoarea += d[i] * rho[i]
            irhoarea += d[i] * irho[i]

            # /* If no more slices or sigma != 0, break immediately */
            i += 1
            if i == n or sigma[i - 1] != 0.0:
                break

            # /* If next slice won't fit, break */
            if rho[i] < rholo:
                rholo = rho[i]
            if rho[i] > rhohi:
                rhohi = rho[i]
            if (rhohi - rholo) * (dz + d[i]) > dA:
                break

            if irho[i] < irholo:
                irholo = irho[i]
            if irho[i] > irhohi:
                irhohi = irho[i]
            if (irhohi - irholo) * (dz + d[i]) > dA:
                break

        # /* dz is only going to be zero if there is a forced break due to
        # * sigma, or if we are accumulating a substrate.  In either case,
        # * we want to accumulate the zero length layer
        # */
        # /* if (dz == 0) continue; */

        # /* Save the layer */
        assert newi < n
        d[newi] = dz
        if i == n:
            # /* printf("contract: adding final sld at %d\n",newi); */
            # /* Last layer uses surface values */
            rho[newi] = rho[n - 1]
            irho[newi] = irho[n - 1]
            # /* No interface for final layer */
        else:
            # /* Middle layers uses average values */
            rho[newi] = rhoarea / dz
            irho[newi] = irhoarea / dz
            sigma[newi] = sigma[i - 1]
        # /* First layer uses substrate values */
        newi += 1

    return newi



# This program is public domain
# Author: Paul Kienzle
"""
Composition space modeling

INCOMPLETE UNUSED UNTESTED CODE


"""

# @PydevCodeAnalysisIgnore

import numpy as np
from numpy import inf, exp
from bumps.parameter import Parameter as Par, to_dict

from .model import Layer
from .materialdb import air


class CompositionSpace(Layer):
    """
    A composition space is much like a stack except that the
    components of the composition space can slide past each
    other, and the parts not taken up by the components is filled
    with a solvent.
    """

    def __init__(self, solvent=air, thickness=0, name=None):
        self.parts = []
        self.solvent = solvent
        if name is None:
            name = "solvent " + solvent.name
        self.name = name
        self.thickness = Par.default(thickness, limits=(0, inf), name=name + " thickness")

    def parameters(self):
        return {
            "solvent": self.solvent.parameters(),
            "parts": [p.parameters() for p in self.parts],
        }

    def to_dict(self):
        return to_dict(
            {
                "type": type(self).__name__,
                "name": self.name,
                "solvent": self.solvent,
                "parts": self.parts,
            }
        )

    def add(self, part=None):
        self.parts.append(part)

    # Array style interface to the parts
    def __getitem__(self, n):
        return self.parts[n]

    def __delitem__(self, n):
        del self.parts[n]

    def __setitem__(self, n, part):
        self.parts[n] = part

    def plot_volume_fraction(self, slabs, ax):
        """
        Composition space items have a plotting routine for showing the
        volume profile.
        """
        # Uniform stepping
        z = np.arange(slabs.dz / 2, self.thickness.value, slabs.dz)

        # Storage for the sub-totals
        volume_total = np.zeros_like(z)

        # Accumulate the parts
        for p in self.parts:
            f = p.profile(z)
            ax.plot(z, f, label=p.name)
            volume_total += f

        # Remainder is solvent
        ax.plot(z, 1 - volume_total, label=self.solvent.name)

    # Render a profile
    def render(self, probe, slabs):
        # Uniform stepping
        z = np.arange(slabs.dz / 2, self.thickness.value, slabs.dz)

        # Storage for the sub-totals
        n, k = len(z), slabs.nprobe
        rho_total = np.zeros((n, k))
        mu_total = np.zeros((n, k))
        volume_total = np.zeros_like(z)

        # Accumulate the parts
        for p in self.parts:
            f, rho, mu = p.f_sld(probe, z)
            rho_total += rho
            mu_total += mu
            volume_total += f

        # Remainder is solvent
        rho, mu = probe.sld(self.solvent)
        rho_total += rho * (1 - volume_total)
        mu_total += mu * (1 - volume_total)

        # Add to model
        w = slabs.dz * np.ones(z.shape)
        slabs.extend(w=w, rho=rho_total, mu=mu_total)


class Part(object):
    def __init__(self, material, profile, fraction=1):
        self.material = material
        self.profile = profile
        self.fraction = Par.default(fraction, limits=(0, 1), name=self.material.name + " fraction")

    def parameters(self):
        return {
            "material": self.material.parameters(),
            "profile": self.profile.parameters(),
            "fraction": self.fraction,
        }

    def to_dict(self):
        return to_dict(
            {
                "type": type(self).__name__,
                "material": self.material,
                "profile": self.profile,
                "fraction": self.fraction,
            }
        )

    def f_sld(self, probe, z):
        # Note: combining f and sld because there my be some
        # composites such as oriented proteins for which the
        # sld and volume change at the same time.
        rho, mu = probe.sld(self.material)
        f = self.fraction.value * self.profile(z)
        return f, rho * f, mu * f


class Gaussian(object):
    """
    Stretched gaussian profile::

        exp ( -(|z-mu|-w)**2 / 2 sigma**2 )   if |z-mu| > w
        1                                     otherwise

    Parameters::

        *center* is mu in (-inf, inf)
        *sigma* is sigma in [0, inf)
        *width* is 2*w in [0, inf)

    Use this profile to represent a flat profile component which is
    subject to gaussian noise in its position in the composition space.

    This profile will be used to construct a volume fraction component
    which a scaling fraction phi and a material describing the scattering
    properties of the profile.
    """

    def __init__(self, center=0, width=0, sigma=1, name="gauss"):
        self.center = Par.default(center, limits=(-inf, inf), name=name + " center")
        self.width = Par.default(width, limits=(0, inf), name=name + " width")
        self.stretch = Par.default(sigma, limits=(0, inf), name=name + " sigma")

    def parameters(self):
        return {"center": self.center, "width": self.width, "sigma": self.stretch}

    def to_dict(self):
        ret = {
            "type": type(self).__name__,
        }
        ret.update(to_dict(self.parameters()))
        return ret

    def __call__(self, z):
        mu, sigma, w = self.center.value, self.stretch.value, self.width.value / 2
        if w <= 0:
            result = exp(-0.5 * ((z - mu) / sigma) ** 2)
        else:
            result = np.ones_like(z)
            idx = abs(z - mu) < w
            result[idx] = exp(-0.5 * ((abs(z[idx] - mu) - w) / sigma) ** 2)
        return result



#!/usr/bin/env python
import sys
import os
from setuptools import setup, Extension

BUILD_EXTENSION = bool(os.environ.get("BUILD_EXTENSION", False)) or "build_ext" in sys.argv


# reflmodule extension
def reflmodule_config():
    S = (
        "reflmodule.cc",
        "methods.cc",
        "reflectivity.cc",
        "magnetic.cc",
        "contract_profile.cc",
        "convolve.cc",
        "convolve_sampled.cc",
        "build_profile.cc",
    )

    Sdeps = ("erf.cc", "methods.h", "rebin.h", "rebin2D.h", "reflcalc.h")
    sources = [os.path.join("refl1d", "lib", "c", f) for f in S]
    depends = [os.path.join("refl1d", "lib", "c", f) for f in Sdeps]
    return Extension(
        "refl1d.reflmodule",
        sources=sources,
        depends=depends,
        language="c++",
    )


dist = setup(
    data_files=[("share/icons", ["extra/refl1d-icon.svg"])],
    ext_modules=[reflmodule_config()] if BUILD_EXTENSION else [],
)
# End of file



"""
Load garefl models into refl1d.

The models themselves don't need to be modified.  See the garefl documentation
for setting up the model.

One extension provided to refl1d that is not available in garefl is the use
of penalty values in the constraints.  The model constraints is able to set::

    fit[0].penalty = FIT_REJECT_PENALTY + distance

Here, *distance* is the distance to the valid region of the search space so
that any fitter that gets lost in a penalty region can more quickly return
to the valid region.  Any penalty value above *FIT_REJECT_PENALTY* will
suppress the evaluation of the model at that point during the fit.

Consider a model with layers (Si | Au | FeNi | air) and the constraint that
d_Au + d_FeNi < 200 A.  The constraints function would be written something
like::

    double excess = fit[0].m.d[1] + fit[0].m.d[2] - 200;
    fit[0].penalty = excess > 0 ? excess*excess+FIT_REJECT_PENALTY : 0.;

Then, if the fit algorithm proposes a value such as Au=125, FeNi=90, the
excess will be 15, and the penalty will be FIT_REJECT_PENALTY+225.

You can use penalties less than *FIT_REJECT_PENALTY*, but these should
correspond to the negative log likelihood of seeing that constraint value
within the model in order for the MCMC uncertainty analysis to work correctly.
*FIT_REJECT_PENALTY* is set to 1e6, which should be high enough that it
doesn't perturb the fit.
"""

__all__ = ["load"]

import sys
import os
from os import getpid
from ctypes import CDLL, c_int, c_double, c_void_p, c_char_p, byref
from threading import current_thread

import numpy as np
from numpy import empty, zeros, array

from bumps.parameter import Parameter, to_dict
from bumps.fitproblem import FitProblem

from .probe import QProbe, PolarizedNeutronQProbe
from .experiment import Experiment
from .model import Stack
from .profile import Microslabs
from .material import SLD, Vacuum

if sys.version_info[0] >= 3:

    def tostr(s):
        return s.decode("ascii")
else:

    def tostr(s):
        return s


def trace(fn):
    """simple function trace function"""
    return fn  # Comment this to turn on tracing

    def wrapper(*args, **kw):
        print("entering %s for thread %s:%s" % (fn.func_name, getpid(), current_thread()))
        ret = fn(*args, **kw)
        print("leaving %s for thread %s:%s" % (fn.func_name, getpid(), current_thread()))
        return ret

    return wrapper


def load(modelfile, probes=None):
    """
    Load a garefl model file as an experiment.

    *modelfile* is a model.so file created from setup.c.

    *probes* is a list of datasets to fit to the models in the model file, or
    None if the model file provides its own data.
    """
    M = experiment(modelfile, probes)
    constraints = M[0]._get_penalty
    if len(M) > 1:
        return FitProblem(M, constraints=constraints)
    else:
        return FitProblem(M[0], constraints=constraints)


def experiment(modelfile, probes=None):
    setup = GareflModel(modelfile)
    if probes:
        if len(probes) != setup.num_models:
            raise ValueError("Number of datasets must match number of models")
        M = [GareflExperiment(setup, k, probe=probes[k]) for k in range(setup.num_models)]
    else:
        M = [GareflExperiment(setup, k) for k in range(setup.num_models)]
    names = setup.par_names()
    low, high = setup.par_bounds()
    value = setup.par_values()
    pars = [Parameter(v, name=s, bounds=(L, H)) for v, s, L, H in zip(value, names, low, high)]
    M[0]._pars = pars
    return M


NOTHING = Vacuum()
NOTHING.name = ""


class GareflExperiment(Experiment):
    def __init__(self, model, index, dz=1, step_interfaces=None, probe=None):
        self.model = model
        self.index = index
        if probe is None:
            probe = model.get_probe(index)
        else:
            model.set_probe(probe)
        self.probe = probe
        self.sample = Stack([NOTHING, NOTHING])
        self.sample[0].interface.fittable = False
        self.step_interfaces = True
        self._slabs = Microslabs(1, dz=dz)
        self._cache = {}  # Cache calculated profiles/reflectivities
        self._pars = None
        self.roughness_limit = 2.35
        self._substrate = SLD(name="substrate", rho=0)
        self._surface = SLD(name="surface", rho=0)
        self._name = None
        self.interpolation = 0

    def parameters(self):
        return self._pars

    def to_dict(self):
        return to_dict(
            {
                "type": type(self).__name__,
                "dll_path": self.model._dll_path,
                "index": self.index,
                "parameters": self.parameters(),
            }
        )

    def _render_slabs(self):
        """
        Build a slab description of the model from the individual layers.
        """
        key = "rendered"
        if key not in self._cache:
            if self._pars is not None:
                pvec = array([p.value for p in self._pars], "d")
                self._chisq = self.model.update_model(pvec, forced=True)

            self._slabs.clear()
            w, rho, irho, rhoM, thetaM = self.model.get_profile(self.index)
            rho, irho, rhoM = 1e6 * rho, 1e6 * irho, 1e6 * rhoM  # remove zeros
            self._slabs.extend(w=w, rho=rho[None, :], irho=irho[None, :])
            # TODO: What about rhoM, thetaM

            # Set values for the Fresnel-normalized reflectivity plot
            self._substrate.rho.value = rho[0]
            self._substrate.irho.value = irho[0]
            self._surface.rho.value = rho[-1]
            self._surface.irho.value = irho[-1]
            self._cache[key] = True
        return self._slabs

    def _get_penalty(self):
        """
        Update the model if necessary and return the penalty value for the point.
        """
        self._render_slabs()
        return self.model.get_penalty()

    def amplitude(self, resolution=True):
        """
        Calculate reflectivity amplitude at the probe points.
        """
        raise NotImplementedError("amplitude not yet available from garefl")

    def reflectivity(self, resolution=True, interpolation=0):
        """
        Calculate predicted reflectivity.
        """
        key = "reflectivity"
        if key not in self._cache:
            self._render_slabs()  # Force recacluation
            if self.probe.polarized:
                Q, Rmm = self.model.get_reflectivity(self.index, 0)
                Q, Rmp = self.model.get_reflectivity(self.index, 1)
                Q, Rpm = self.model.get_reflectivity(self.index, 2)
                Q, Rpp = self.model.get_reflectivity(self.index, 3)
                self._cache[key] = Q, (Rmm, Rmp, Rpm, Rpp)
            else:
                Q, R = self.model.get_reflectivity(self.index, 0)
                self._cache[key] = Q, R
        return self._cache[key]

    def output_model(self):
        self.model.output_model()


class GareflModel(object):
    def __init__(self, path):
        self._dll_path = os.path.abspath(path)
        self._load_dll()
        self._setup_model()

    @trace
    def _load_dll(self):
        dll = CDLL(self._dll_path)
        dll.ex_get_data.restype = c_char_p
        dll.ex_set_data.restype = c_int
        dll.setup_models.restype = c_void_p
        dll.ex_par_name.restype = c_char_p
        dll.ex_get_penalty.restype = c_double
        self.dll = dll
        self.num_models = 0

    @trace
    def _setup_model(self):
        if self.num_models:
            raise RuntimeError("Model already loaded")
        MODELS = c_int()
        self.models = c_void_p(self.dll.setup_models(byref(MODELS)))
        self.num_models = MODELS.value
        self.num_pars = self.dll.ex_npars(self.models)
        lo, hi = self._par_bounds()
        small = np.max(np.vstack((abs(lo), abs(hi))), axis=0) < 1e-3
        self.scale = np.where(small, 1e6, 1)

        # TODO: better way to force recalc on load
        self.update_model(self.par_values())

    # Pickle protocol doesn't support ctypes linkage; reload the
    # module on the other side.
    def __getstate__(self):
        return self._dll_path

    def __setstate__(self, state):
        self._dll_path = state
        self._load_dll()
        self._setup_model()

    def clear_model(self):
        if self.num_models:
            self.dll.ex_fit_destroy(self.models)
            self.num_models = 0

    @trace
    def update_model(self, p, weighted=1, approximate_roughness=0, forced=False):
        p = p / self.scale
        # assert p.flags.aligned and (p.flags.c_contiguous or p.flags.f_contiguous)
        # assert p.size == self.num_pars
        self.dll.ex_set_pars(self.models, p.ctypes)
        chisq = self.dll.ex_update_models(self.models, self.num_models, weighted, approximate_roughness, int(forced))
        return chisq

    @trace
    def get_probe(self, k):
        """
        Return a probe for an individual garefl model.
        """
        data = [self._getdata(k, xs) for xs in range(4)]
        if any(d is not None for d in data[1:]):
            probe = PolarizedNeutronQProbe(xs=data)
        else:
            probe = data[0]
        return probe

    def _getdata(self, k, xs):
        """
        Convert a single model into a probe
        """
        n = self.dll.ex_ndata(self.models, k, xs)
        if n == 0:
            return None
        data = empty((n, 4), "d")
        filename = tostr(self.dll.ex_get_data(self.models, k, xs, data.ctypes))
        Q, dQ, R, dR = data.T
        probe = QProbe(Q, dQ, data=(R, dR), name=filename)
        return probe

    @trace
    def set_probe(self, k, probe):
        """
        Return a probe for an individual garefl model.
        """
        if probe.polarized:
            for xs, probe_xs in enumerate(probe.xs):
                self._setdata(k, xs, probe_xs)
        else:
            self._setdata(k, 0, probe)

    def _setdata(self, k, xs, probe):
        if probe is not None:
            n = probe.Q
            data = empty((n, 4), dtype="d", order="F")
            data[:, 0] = probe.Q
            data[:, 1] = probe.dQ
            data[:, 2] = probe.R
            data[:, 3] = probe.dR
        else:
            n = 0
            data = empty((n, 4), "d")
        result = self.dll.ex_set_data(self.models, k, xs, n, data.ctypes)
        if result < 0:
            raise RuntimeError("unable to create data in garefl")

    @trace
    def get_profile(self, k):
        n = self.dll.ex_nprofile(self.models, k)
        w, rho, irho, rhoM, thetaM = [zeros(n, "d") for _ in range(5)]
        self.dll.ex_get_profile(self.models, k, w.ctypes, rho.ctypes, irho.ctypes, rhoM.ctypes, thetaM.ctypes)
        return w[::-1], rho[::-1], irho[::-1], rhoM[::-1], thetaM[::-1]

    @trace
    def get_reflectivity(self, k, xs):
        n = self.dll.ex_ncalc(self.models, k)
        Q, R = empty(n, "d"), empty(n, "d")
        self.dll.ex_get_reflectivity(self.models, k, xs, Q.ctypes, R.ctypes)
        return Q, R

    @trace
    def get_penalty(self):
        # print "penalty", self.dll.ex_get_penalty(self.models), self.par_values()
        return self.dll.ex_get_penalty(self.models)

    @trace
    def output_model(self):
        """
        Run the output_model function
        """
        return self.dll.ex_output_model(self.models)

    @trace
    def par_names(self):
        return [tostr(self.dll.ex_par_name(self.models, i)) for i in range(self.num_pars)]

    @trace
    def par_bounds(self):
        return self._par_bounds() * self.scale

    @trace
    def _par_bounds(self):
        lo, hi = empty(self.num_pars, "d"), empty(self.num_pars, "d")
        self.dll.ex_par_bounds(self.models, lo.ctypes, hi.ctypes)
        return lo, hi

    @trace
    def par_values(self):
        p = empty(self.num_pars, "d")
        self.dll.ex_par_values(self.models, p.ctypes)
        return p * self.scale



"""
Freeform modeling with B-Splines

**DEPRECATED** Use refl1d.mono instead.
"""

import numpy as np
from numpy import inf
from bumps.parameter import Parameter as Par, to_dict
from bumps.bspline import pbs, bspline

from .model import Layer
from . import util


# TODO: add left_sld, right_sld to all layers so that fresnel works
# TODO: access left_sld, right_sld so freeform doesn't need left, right
# TODO: restructure to use vector parameters
# TODO: allow the number of layers to be adjusted by the fit
class FreeLayer(Layer):
    """
    A freeform section of the sample modeled with B-splines.

    sld (rho) and imaginary sld (irho) can be modeled with a separate
    number of control points. The control points can be equally spaced
    in the layers unless rhoz or irhoz are specified. If the z values
    are given, they must be in the range [0, 1].  One control point is
    anchored at either end, so there are two fewer z values than controls
    if z values are given.

    Layers have a slope of zero at the ends, so the automatically blend
    with slabs.
    """

    def __init__(self, thickness=0, left=None, right=None, rho=(), irho=(), rhoz=(), irhoz=(), name="Freeform"):
        self.name = name
        self.left, self.right = left, right
        self.thickness = Par.default(thickness, limits=(0, inf), name=name + " thickness")
        self.rho, self.irho, self.rhoz, self.irhoz = [
            [Par.default(p, name=name + " [%d] %s" % (i, part), limits=limits) for i, p in enumerate(v)]
            for v, part, limits in zip(
                (rho, irho, rhoz, irhoz), ("rho", "irho", "rhoz", "irhoz"), ((-inf, inf), (-inf, inf), (0, 1), (0, 1))
            )
        ]
        if len(self.rhoz) > 0 and len(self.rhoz) != len(self.rho):
            raise ValueError("must have one z value for each rho")
        if len(self.irhoz) > 0 and len(self.irhoz) != len(self.irho):
            raise ValueError("must have one z value for each irho")

    def parameters(self):
        return {
            "rho": self.rho,
            "rhoz": self.rhoz,
            "irho": self.irho,
            "irhoz": self.irhoz,
            # TODO: left/right pars are already listed in the stack
            "left": self.left.parameters(),
            "right": self.right.parameters(),
        }

    def to_dict(self):
        return to_dict(
            {
                "type": type(self).__name__,
                "name": self.name,
                "parameters": self.parameters(),
            }
        )

    def render(self, probe, slabs):
        thickness = self.thickness.value
        left_rho, left_irho = self.left.sld(probe)
        right_rho, right_irho = self.right.sld(probe)
        Pw, Pz = slabs.microslabs(thickness)
        t = Pz / thickness
        Prho = _profile(left_rho, right_rho, self.rho, self.rhoz, t)
        Pirho = _profile(left_irho, right_irho, self.irho, self.irhoz, t)
        slabs.extend(rho=[Prho], irho=[Pirho], w=Pw)


class FreeformInterface01(Layer):
    """
    A freeform section of the sample modeled with B-splines.

    sld (rho) and imaginary sld (irho) can be modeled with a separate
    number of control points. The control points can be equally spaced
    in the layers unless rhoz or irhoz are specified. If the z values
    are given, they must be in the range [0, 1].  One control point is
    anchored at either end, so there are two fewer z values than controls
    if z values are given.

    Layers have a slope of zero at the ends, so the automatically blend
    with slabs.
    """

    def __init__(self, thickness=0, interface=0, below=None, above=None, z=None, vf=None, name="Interface"):
        self.name = name
        self.below, self.above = below, above
        self.thickness = Par.default(thickness, limits=(0, inf), name=name + " thickness")
        self.interface = Par.default(interface, limits=(0, inf), name=name + " interface")
        if len(z) != len(vf):
            raise ValueError("Need one vf for every z")
        # if len(z) != len(vf)+2:
        #    raise ValueError("Only need vf for interior z, so len(z)=len(vf)+2")
        self.z = [Par.default(p, name=name + " z[%d]" % i, limits=(0, 1)) for i, p in enumerate(z)]
        self.vf = [Par.default(p, name=name + " vf[%d]" % i, limits=(0, 1)) for i, p in enumerate(vf)]

    def parameters(self):
        return {
            "thickness": self.thickness,
            "interface": self.interface,
            "z": self.z,
            "vf": self.vf,
            "below": self.below.parameters(),
            "above": self.above.parameters(),
        }

    def to_dict(self):
        return to_dict(
            {
                "type": type(self).__name__,
                "name": self.name,
                "parameters": self.parameters(),
            }
        )

    def render(self, probe, slabs):
        thickness = self.thickness.value
        left_rho, left_irho = self.below.sld(probe)
        right_rho, right_irho = self.above.sld(probe)
        z = np.hstack((0, sorted([v.value for v in self.z]), 1))
        vf = np.hstack((0, sorted([v.value for v in self.vf]), 1))
        Pw, Pz = slabs.microslabs(thickness)
        t = Pz / thickness
        offset, profile = pbs(z, vf, t, parametric=False, clamp=True)
        Pw, profile = util.merge_ends(Pw, profile, tol=1e-3)
        Prho = (1 - profile) * left_rho + profile * right_rho
        Pirho = (1 - profile) * left_irho + profile * right_irho
        slabs.extend(rho=[Prho], irho=[Pirho], w=Pw)


class FreeInterface(Layer):
    """
    A freeform section of the sample modeled with monotonic splines.

    Layers have a slope of zero at the ends, so the automatically blend
    with slabs.
    """

    def __init__(self, interface=0, below=None, above=None, dz=None, dp=None, name="Interface"):
        self.name = name
        self.below, self.above = below, above
        self.interface = Par.default(interface, limits=(0, inf), name=name + " interface")

        # Choose reasonable defaults if not given
        if dp is None and dz is None:
            dp = [1] * 5
        if dp is None:
            dp = [1] * len(dz)
        if dz is None:
            dz = [10.0 / len(dp)] * len(dp)
        if len(dz) != len(dp):
            raise ValueError("Need one dz for every dp")
        # if len(z) != len(vf)+2:
        #    raise ValueError("Only need vf for interior z, so len(z)=len(vf)+2")
        self.dz = [Par.default(p, name=name + " dz[%d]" % i, limits=(0, inf)) for i, p in enumerate(dz)]
        self.dp = [Par.default(p, name=name + " dp[%d]" % i, limits=(0, inf)) for i, p in enumerate(dp)]

    def _get_thickness(self):
        w = sum(v.value for v in self.dz)
        return Par(w, name=self.name + " thickness")

    def _set_thickness(self, v):
        if v != 0:
            raise ValueError("thickness cannot be set for FreeInterface")

    thickness = property(_get_thickness, _set_thickness)

    def parameters(self):
        return {
            "interface": self.interface,
            "below": self.below.parameters(),
            "above": self.above.parameters(),
            "dz": self.dz,
            "dp": self.dp,
        }

    def to_dict(self):
        return to_dict(
            {
                "type": type(self).__name__,
                "name": self.name,
                "parameters": self.parameters(),
            }
        )

    def render(self, probe, slabs):
        left_rho, left_irho = self.below.sld(probe)
        right_rho, right_irho = self.above.sld(probe)
        z = np.hstack((0, np.cumsum([v.value for v in self.dz])))
        p = np.hstack((0, np.cumsum([v.value for v in self.dp])))
        if p[-1] == 0:
            p[-1] = 1
        p /= p[-1]
        Pw, Pz = slabs.microslabs(z[-1])
        _, profile = pbs(z, p, Pz, parametric=False, clamp=True)
        profile = np.clip(profile, 0, 1)
        Pw, profile = util.merge_ends(Pw, profile, tol=1e-3)
        Prho = (1 - profile) * left_rho + profile * right_rho
        Pirho = (1 - profile) * left_irho + profile * right_irho
        slabs.extend(rho=[Prho], irho=[Pirho], w=Pw)


def _profile(left, right, control, z, t):
    cy = np.hstack((left, [p.value for p in control], right))
    if len(z) > 0:
        cx = np.hstack((0, [p.value for p in z], 1))
        return pbs(cx, cy, t)
    else:
        return bspline(cy, t)



#!/usr/bin/env python

# Copyright (C) 2006-2011, University of Maryland
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/ or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# Author: James Krycka, Paul Kienzle
"""
Start Refl1D with the graphical user interface.
"""

# ========================== Start program ====================================
if __name__ == "__main__":
    # This is necessary when running the application from a frozen image and
    # using the --parallel option.  Note that freeze_support() has no effect
    # when running from a python script (i.e., in a non-frozen environment).
    import multiprocessing

    multiprocessing.freeze_support()

    import refl1d.main

    refl1d.main.gui()



# DEAD CODE
class Vector(object):
    """
    Labelled set of numbers, possibly with uncertainty.

    This will usually be used for graphing.
    """

    def __init__(self, data=(), err=(), name="", units=""):
        self.name = name
        self.units = units
        self.data = data
        self.err = err

    def plot(self, page, **kw):
        return page.bar(self, **kw)


class Data(object):
    """
    A simple 1-D dataset x vs. y.
    """

    def __init__(self, x, y):
        self.x, self.y = x, y

    def plot(self, page, **kw):
        return page.marker(x, y, **kw)


class Trend(object):
    """
    A line fx vs. fy.  Note that this may be infinite extent, and
    should probably have a mechanism for recalculating given new
    axes limits.
    """

    def __init__(self, x=None, y=None):
        self.fx, self.fy = fx, fy

    def plot(self, page, **kw):
        return page.line(fx, fy, **kw)


class Histogram(object):
    def __init__(self, x=None):
        self.x = x

    def plot(self, page, **kw):
        return huh


class Fit(object):
    """
    Data plus trend.
    """

    def __init__(self, data=None, trend=None):
        self.data, self.trend = data, trend

    def plot(self, page):
        h1 = self.data.plot(page)
        h2 = self.trend.plot(page, style=h.style.complement())
        return huh


class Data2D(object):
    def __init__(self, x, y, z):
        self.x, self.y, self.z = x, y, z

    def plot(self, page):
        page.surface(x, y, z)


# Colors need to complement or contrast with existing colors.
class Style(object):
    def complement(self, k=1, n=1):
        return Complement(self, k, n)

    def contrast(self):
        return Contrast(self)


class Complement(Style):
    def __init__(self, base, k, n):
        self.base, self.k, self.n = base, k, n

    def color(self):
        return self._color.darken(k, n)


class Contrast(Style):
    def __init__(self, base, k, n):
        self.base, self.k, self.n = base, k, n

    def color(self):
        return palette.next_color()


class Page(object):
    def marker(self, x=None, y=None, **kw):
        """
        Add markers for x, y to the axes.  If x, y have error bars, these
        will be included.  Legends are automatic.

        Returns a handle.
        """

    def line(self, x=None, y=None, **kw):
        """
        Add lines for x, y to the axes.  If y has error bars, these will
        be drawn appropriately.  Uncertainty in x will be ignored.

        Returns a handle.
        """

    def surface(self, x=None, y=None, z=None, **kw):
        """
        Add a surface to the axes, with appropriate representation for
        uncertainty in x, y and z.

        Returns a handle.
        """

    def labels(self, x, y, text, **kw):
        """
        Add labels to the axes.
        """


class MplPage(Page):
    def _new_axes(self, x, y):
        pass

    def _find_axes(self, x, y):
        pass

    def _get_axes(self, x, y):
        ax = self._find_axes(x, y)
        if ax is None:
            ax = self._new_axes(x, y)
        return ax

    def marker(self, x=None, y=None, **kw):
        ax = self._get_axes(x, y)
        ax.plot(x.data, y.data)

    def line(self, x=None, y=None, **kw):
        ax = self._get_axes(x, y)
        ax.plot(x.data, y.data, "-")

    def surface(self, x=None, y=None, z=None, **kw):
        ax = self._get_axes(x, y)
        ax.pcolormesh(x.data, y.data, z.data)



# This program is public domain
# Author: Paul Kienzle
r"""
Magnetic modeling for 1-D reflectometry.

Magnetic properties are tied to the structural description of the
but only loosely.

There may be dead regions near the interfaces of magnetic materials.

Magnetic behaviour may be varying in complex ways within and
across structural boundaries.  For example, the ma
Indeed, the pattern may continue
across spacer layers, going to zero in the magnetically dead
region and returning to its long range variation on entry to
the next magnetic layer.  Magnetic multilayers may exhibit complex
magnetism throughout the repeated section while the structural
components are fixed.

The scattering behaviour is dependent upon net field strength relative to
polarization direction.   This arises from three underlying quantities:
the strength of the individual dipole moments in the layer, the degree
of alignment of these moments, and the net direction of the alignment.
The strength of the dipole moment depends on the details of the electronic
structure, so unlike the nuclear scattering potential, it cannot be readily
determined from material composition.  Similarly, net magnetization
depends on the details of the magnetic domains within the material, and
cannot readily be determined from first principles.  The interaction
potential of the net magnetic moment depends on the alignment of the field
with respect to the beam, with a net scattering length density of
:math:`\rho_M \cos(\theta_M)`.  Clearly the scattering measurement will
not be able to distinguish between a reduced net magnetic strength
:math:`\rho_M` and a change in orientation :math:`\theta_M` for an
individual measurement, as should be apparent from the correlated
uncertainty plot produced when both parameters are fit.

Magnetism support is split into two parts: describing the layers
and anchoring them to the structure.
"""

from __future__ import print_function

from dataclasses import dataclass, field
from typing import Optional, Any, Union, Dict, Callable, Literal, Tuple, List, Literal

import numpy as np
from numpy import inf
from bumps.parameter import Parameter, flatten, to_dict
from bumps.mono import monospline

from .reflectivity import BASE_GUIDE_ANGLE as DEFAULT_THETA_M


@dataclass
class BaseMagnetism:
    """
    Magnetic properties of the layer.

    Magnetism is attached to set of nuclear layers by setting the
    *magnetism* property of the first layer to the rendered for the
    magnetic profile, and setting *extent* to the number of nuclear
    layers attached to the magnetism object.

    *dead_below* and *dead_above* are dead regions within the magnetic
    extent, which allow you to shift the magnetic interfaces relative
    to the nuclear interfaces.

    *interface_below* and *interface_above* are the interface widths
    for the magnetic layer, which default to the interface widths for
    the corresponding nuclear layers if no interfaces are specified.
    For consecutive layers, only *interface_above* is used; any value
    for *interface_below* is ignored.
    """

    name: str
    extent: float
    dead_below: Union[Parameter, Literal[None]]
    dead_above: Union[Parameter, Literal[None]]
    interface_below: Union[Parameter, Literal[None]]
    interface_above: Union[Parameter, Literal[None]]

    def __init__(self, extent=1, dead_below=0, dead_above=0, interface_below=None, interface_above=None, name="LAYER"):
        self.dead_below = Parameter.default(dead_below, limits=(0, None), name=name + " deadM below")
        self.dead_above = Parameter.default(dead_above, limits=(0, None), name=name + " deadM above")
        if interface_below is not None:
            interface_below = Parameter.default(interface_below, limits=(0, None), name=name + " interfaceM below")
        if interface_above is not None:
            interface_above = Parameter.default(interface_above, limits=(0, None), name=name + " interfaceM above")
        self.interface_below = interface_below
        self.interface_above = interface_above
        self.name = name
        self.extent = extent

    def parameters(self) -> Dict[str, Union[Parameter, List[Parameter], Literal[None]]]:
        return {
            "dead_below": self.dead_below,
            "dead_above": self.dead_above,
            "interface_below": self.interface_below,
            "interface_above": self.interface_above,
        }

    def to_dict(self):
        return to_dict(
            {
                "type": type(self).__name__,
                "name": self.name,
                "extent": self.extent,
                "dead_below": self.dead_below,
                "dead_above": self.dead_above,
                "interface_below": self.interface_below,
                "interface_above": self.interface_above,
            }
        )

    def set_layer_name(self, name):
        """
        Update the names of the magnetic parameters with the name of the
        layer if it has not already been set.  This is necessary since we don't
        know the layer name until after we have constructed the magnetism object.
        """
        if self.name == "LAYER":
            for p in flatten(self.parameters()):
                p.name = p.name.replace("LAYER", name)
            self.name = name


@dataclass(init=False)
class Magnetism(BaseMagnetism):
    """
    Region of constant magnetism.

    *rhoM* is the magnetic SLD the layer. Default is :code:`rhoM=0`.

    *thetaM* is the magnetic angle for the layer. Default is :code:`thetaM=270`.

    *name* is the base name for the various layer parameters.

    *extent* defines the number of nuclear layers covered by the magnetic layer.

    *dead_above* and *dead_below* define magnetically dead layers at the
    nuclear boundaries.  These can be negative if magnetism extends beyond
    the nuclear boundary.

    *interface_above* and *interface_below* define the magnetic interface
    at the boundaries, if it is different from the nuclear interface.
    """

    rhoM: Parameter
    thetaM: Parameter

    def __init__(self, rhoM: Union[float, Parameter] = 0, thetaM=DEFAULT_THETA_M, name="LAYER", **kw):
        BaseMagnetism.__init__(self, name=name, **kw)
        self.rhoM = Parameter.default(rhoM, name=name + " rhoM")
        self.thetaM = Parameter.default(thetaM, limits=(0, 360), name=name + " thetaM")

    def parameters(self):
        parameters = BaseMagnetism.parameters(self)
        parameters.update(rhoM=self.rhoM, thetaM=self.thetaM)
        return parameters

    def to_dict(self):
        result = BaseMagnetism.to_dict(self)
        result["rhoM"] = to_dict(self.rhoM)
        result["thetaM"] = to_dict(self.thetaM)
        return result

    def render(self, probe, slabs, thickness, anchor, sigma):
        slabs.add_magnetism(
            anchor=anchor, w=[thickness], rhoM=[self.rhoM.value], thetaM=[self.thetaM.value], sigma=sigma
        )

    def __str__(self):
        return "magnetism(%g)" % self.rhoM.value

    def __repr__(self):
        return "Magnetism(rhoM=%g, thetaM=%g)" % (self.rhoM.value, self.thetaM.value)


@dataclass(init=False)
class MagnetismStack(BaseMagnetism):
    """
    Magnetic slabs within a magnetic layer.

    *weight* is the relative thickness of each layer relative to the nuclear
    stack to which it is anchored.  Weights are automatically normalized to 1.
    Default is :code:`weight=[1]` equal size layers.

    *rhoM* is the magnetic SLD for each layer. Default is :code:`rhoM=[0]`
    for shared magnetism in all the layers.

    *thetaM* is the magnetic angle for each layer.  Default is
    :code:`thetaM=[270]` for no magnetic twist.

    **Not yet implemented.**
    *interfaceM* is the magnetic interface for all but the last layer.  Default
    is :code:`interfaceM=[0]` for equal width interfaces in all layers.

    *name* is the base name for the various layer parameters.

    *extent* defines the number of nuclear layers covered by the magnetic layer.

    *dead_above* and *dead_below* define magnetically dead layers at the
    nuclear boundaries.  These can be negative if magnetism extends beyond
    the nuclear boundary.

    *interface_above* and *interface_below* define the magnetic interface
    at the boundaries, if it is different from the nuclear interface.
    """

    weight: List[Parameter]
    rhoM: List[Parameter]
    thetaM: List[Parameter]

    def __init__(self, weight=None, rhoM=None, thetaM=None, interfaceM=None, name="LAYER", **kw):
        weight_n = 0 if weight is None else len(weight)
        rhoM_n = 0 if rhoM is None else len(rhoM)
        thetaM_n = 0 if thetaM is None else len(thetaM)
        interfaceM_n = 0 if interfaceM is None else (len(interfaceM) + 1)
        n = max(weight_n, rhoM_n, thetaM_n, interfaceM_n)

        if n == 0:
            raise ValueError("Must specify one of weight, rhoM, thetaM or interfaceM as vector")
        if (
            (weight_n > 1 and weight_n != n)
            or (rhoM_n > 1 and rhoM_n != n)
            or (thetaM_n > 1 and thetaM_n != n)
            or (interfaceM_n > 1 and interfaceM_n != n)
        ):
            raise ValueError("Inconsistent lengths for weight, rhoM, thetaM and interfaceM")

        # TODO: intefaces need to be implemented in profile.add_magnetism
        if interfaceM is not None:
            raise NotImplementedError("Doesn't yet support magnetic interfaces")

        if weight is None:
            weight = [1]
        if rhoM is None:
            rhoM = [0]
        if thetaM is None:
            thetaM = [DEFAULT_THETA_M]
        # if interfaceM is None:
        #    interfaceM = [0]

        BaseMagnetism.__init__(self, name=name, **kw)
        self.weight = [Parameter.default(v, name=name + " weight[%d]" % i) for i, v in enumerate(weight)]
        self.rhoM = [Parameter.default(v, name=name + " rhoM[%d]" % i) for i, v in enumerate(rhoM)]
        self.thetaM = [Parameter.default(v, name=name + " thetaM[%d]" % i) for i, v in enumerate(thetaM)]
        # self.interfaceM = [Parameter.default(v, name=name+" interfaceM[%d]"%i)
        #                   for i, v in enumerate(interfaceM)]

    def parameters(self):
        parameters = BaseMagnetism.parameters(self)
        parameters.update(
            rhoM=self.rhoM,
            thetaM=self.thetaM,
            # interfaceM=self.interfaceM,
            weight=self.weight,
        )
        return parameters

    def to_dict(self):
        result = BaseMagnetism.to_dict(self)
        result["weight"] = to_dict(self.weight)
        result["rhoM"] = to_dict(self.rhoM)
        result["thetaM"] = to_dict(self.thetaM)
        # result['interfaceM'] = to_dict(self.interfaceM)
        return result

    def render(self, probe, slabs, thickness, anchor, sigma):
        w = np.array([p.value for p in self.weight])
        w *= thickness / np.sum(w)
        rhoM = [p.value for p in self.rhoM]
        thetaM = [p.value for p in self.thetaM]
        # interfaceM = [p.value for p in self.interfaceM]
        if len(rhoM) == 1:
            rhoM = [rhoM[0]] * len(w)
        if len(thetaM) == 1:
            thetaM = [thetaM[0]] * len(w)
        # if len(interfaceM) == 1:
        #    interfaceM = [interfaceM[0]]*(len(w)-1)

        slabs.add_magnetism(
            anchor=anchor,
            w=w,
            rhoM=rhoM,
            thetaM=thetaM,
            # interfaceM=interfaceM,
            sigma=sigma,
        )

    def __str__(self):
        return "MagnetismStack(%d)" % (len(self.rhoM))

    def __repr__(self):
        return "MagnetismStack"


@dataclass(init=False)
class MagnetismTwist(BaseMagnetism):
    """
    Linear change in magnetism throughout layer.

    *rhoM* contains the *(left, right)* values for the magnetic scattering
    length density.  The number of steps is determined by the model *dz*.

    *thetaM* contains the *(left, right)* values for the magnetic angle.

    *name* is the base name for the various layer parameters.

    *extent* defines the number of nuclear layers covered by the magnetic layer.

    *dead_above* and *dead_below* define magnetically dead layers at the
    nuclear boundaries.  These can be negative if magnetism extends beyond
    the nuclear boundary.

    *interface_above* and *interface_below* define the magnetic interface
    at the boundaries, if it is different from the nuclear interface.
    """

    rhoM: List[Parameter]
    thetaM: List[Parameter]

    magnetic = True

    def __init__(self, rhoM=(0, 0), thetaM=(DEFAULT_THETA_M, DEFAULT_THETA_M), name="LAYER", **kw):
        BaseMagnetism.__init__(self, name=name, **kw)
        self.rhoM = [Parameter.default(v, name=name + " rhoM[%d]" % i) for i, v in enumerate(rhoM)]
        self.thetaM = [Parameter.default(v, name=name + " thetaM[%d]" % i) for i, v in enumerate(thetaM)]

    def parameters(self):
        parameters = BaseMagnetism.parameters(self)
        parameters.update(rhoM=self.rhoM, thetaM=self.thetaM)
        return parameters

    def to_dict(self):
        result = BaseMagnetism.to_dict(self)
        result["rhoM"] = to_dict(self.rhoM)
        result["thetaM"] = to_dict(self.thetaM)
        return result

    def render(self, probe, slabs, thickness, anchor, sigma):
        w, z = slabs.microslabs(thickness)
        rhoM = np.linspace(self.rhoM[0].value, self.rhoM[1].value, len(z))
        thetaM = np.linspace(self.thetaM[0].value, self.thetaM[1].value, len(z))
        slabs.add_magnetism(anchor=anchor, w=w, rhoM=rhoM, thetaM=thetaM, sigma=sigma)

    def __str__(self):
        return "twist(%g->%g)" % (self.rhoM[0].value, self.rhoM[1].value)

    def __repr__(self):
        return "MagneticTwist"


@dataclass(init=False)
class FreeMagnetism(BaseMagnetism):
    """
    Spline change in magnetism throughout layer.

    Defines monotonic splines for rhoM and thetaM with shared knot positions.

    *z* is position of the knot in [0, 1] relative to the magnetic layer
    thickness.  The *z* coordinates are automatically sorted before
    rendering, leading to multiple equivalent solutions if knots are swapped.

    *rhoM* gives the magnetic scattering length density for each knot.

    *thetaM* gives the magnetic angle for each knot.

    *name* is the base name for the various layer parameters.

    *dead_above* and *dead_below* define magnetically dead layers at the
    nuclear boundaries.  These can be negative if magnetism extends beyond
    the nuclear boundary.

    *interface_above* and *interface_below* define the magnetic interface
    at the boundaries, if it is different from the nuclear interface.
    """

    z: List[Parameter]
    rhoM: List[Parameter]
    thetaM: List[Parameter]

    magnetic = True

    def __init__(self, z=(), rhoM=(), thetaM=(), name="LAYER", **kw):
        BaseMagnetism.__init__(self, name=name, **kw)

        def parvec(vector, name, limits):
            return [Parameter.default(p, name=name + "[%d]" % i, limits=limits) for i, p in enumerate(vector)]

        self.rhoM, self.thetaM, self.z = [
            parvec(v, name + " " + part, limits)
            for v, part, limits in zip((rhoM, thetaM, z), ("rhoM", "thetaM", "z"), ((0, None), (0, 360), (0, 1)))
        ]
        if len(self.z) != len(self.rhoM):
            raise ValueError("must have one position z for each rhoM")
        if len(self.thetaM) > 0 and len(self.rhoM) != len(self.thetaM):
            raise ValueError("must have one thetaM for each rhoM")

    def parameters(self):
        parameters = BaseMagnetism.parameters(self)
        parameters.update(rhoM=self.rhoM, thetaM=self.thetaM, z=self.z)
        return parameters

    def to_dict(self):
        result = BaseMagnetism.to_dict(self)
        result["z"] = to_dict(self.z)
        result["rhoM"] = to_dict(self.rhoM)
        result["thetaM"] = to_dict(self.thetaM)
        return result

    def profile(self, Pz, thickness):
        mbelow, tbelow = 0, (self.thetaM[0].value if self.thetaM else DEFAULT_THETA_M)
        mabove, tabove = 0, (self.thetaM[-1].value if self.thetaM else DEFAULT_THETA_M)
        z = np.sort([0.0] + [p.value for p in self.z] + [1.0]) * thickness

        rhoM = np.hstack((mbelow, [p.value for p in self.rhoM], mabove))
        PrhoM = monospline(z, rhoM, Pz)

        if np.any(np.isnan(PrhoM)):
            print("in mono with bad PrhoM")
            print("z %s" % str(z))
            print("p %s" % str([p.value for p in self.z]))

        if len(self.thetaM) > 1:
            thetaM = np.hstack((tbelow, [p.value for p in self.thetaM], tabove))
            PthetaM = monospline(z, thetaM, Pz)
        elif len(self.thetaM) == 1:
            PthetaM = self.thetaM[0].value * np.ones_like(PrhoM)
        else:
            PthetaM = DEFAULT_THETA_M * np.ones_like(PrhoM)
        return PrhoM, PthetaM

    def render(self, probe, slabs, thickness, anchor, sigma):
        Pw, Pz = slabs.microslabs(thickness)
        rhoM, thetaM = self.profile(Pz, thickness)
        slabs.add_magnetism(anchor=anchor, w=Pw, rhoM=rhoM, thetaM=thetaM, sigma=sigma)

    def __str__(self):
        return "freemag(%d)" % (len(self.rhoM))

    def __repr__(self):
        return "FreeMagnetism"
